'''
This module contains helper functions for the model classes.
'''

from __future__ import(
	division,
	print_function,
	)

__docformat__ = 'restructuredtext en'
__all__ = ['_bd_calc_A',
			'_calc_ghat', 
			'_calc_p', 
			'_calc_R',
			'_rpo_calc_A'
			]

import numpy as np

from numpy.linalg import norm
from scipy.optimize import nnls

#import helper functions
from .core_functions import(
	assert_len,
	)

#define function to calculte the A matrix for DAEM models
def _bd_calc_A(k, t, logged = False):
	'''
	Calculates the A matrix for a BioDecay model (e.g. an IsoCaRB run).

	Parameters
	----------
	k : array-like
		Array of first-order rate constants to be used in the A matrix, in s-1.
		Length `nk`.

	t : array-like
		Array of timepoints to be used in the A matrix, in seconds. Length `nt`.

	logged : boolean
		If `True`, treats inputted `k` array as the natural log of k (i.e. lambda
		in Forney and Rothman, 2012 notation). If `False`, treats `k` as a
		linear array. Defaults to `False`.

	Returns
	-------
	A : np.ndarray
		2d array of the Laplace transform for the first-order decay model. 
		Shape [`nt` x `nk`].

	References
	----------
	[1] B.P. Boudreau and B.R. Ruddick (1991) On a reactive continuum
		representation of organic matter diagenesis. *Am. J. Sci.*, **291**,
		507-538.

	[2] D.C. Forney and D.H. Rothman (2012) Inverse method for calculating
		respiration rates from decay time series. *Biogeosciences*, **9**,
		3601-3612.
	'''

	#set constants
	nt = len(t)
	nk = len(k)

	#get arrays in the right format and ensure lengths
	k = assert_len(k, nk) #s-1
	t = assert_len(t, nt) #s

	#calculate k gradient
	dk = np.gradient(k)

	#pre-allocate A
	A = np.zeros([nt,nk])

	#generate t and k matrices
	tmat = np.outer(t, np.ones(nk))
	kmat = np.outer(np.ones(nt), k)

	#generate and dk matrix
	dkmat = np.outer(np.ones(nt), dk)

	if logged is True:
		#calculate A matrix
		A = np.exp(-np.exp(kmat)*tmat)*dkmat #s-1

	else:
		#calculate A matrix
		A = np.exp(-kmat*tmat)*dkmat #s-1

	return A

#define a function to generate estimated time data from model and ratedata
def _calc_ghat(model, ratedata):
	'''
	Calculates the timedata for a given ``rp.RateData`` and ``rp.Model`` 
	instance.

	Parameters
	----------
	model : rp.Model
		The ``rp.Model`` instance used to calculate the forward model.

	ratedata : rp.RateData
		The ``rp.RateData`` instance containing the reactive continuum data.

	Returns
	-------
	ghat : array-like
		Array of estimated fraction of total carbon remaining at each 
		timestep. Length `nt`.
	'''

	return np.inner(model.A, ratedata.p)

#define a function to generate estimated rate data from model and timedata
def _calc_p(model, timedata, lam):
	'''
	Calculates the reactive continuum of rates (or E, for DAEM) for a given
	``rp.TimeData`` and ``rp.Model`` instance.

	Parameters
	----------
	model : rp.Model
		``rp.Model`` instance containing the A matrix to use for calculation.

	timedata : rp.TimeData
		``rp.Timedata`` instance containing the fraction remaining with time 
		array to use for the calculation.

	lam : scalar
		Tikhonov regularization weighting factor, `lambda`.

	Returns
	-------
	p : np.ndarray
		Array of the pdf of the discretized distribution of rates (or E, for 
		DAEM).

	resid : float
		Residual RMSE between true and modeled time data.

	rgh : float
		Roughness RMSE from Tikhonov Regularization.

	References
	----------
	[1] D.C. Forney and D.H. Rothman (2012) Inverse method for calculating
		respiration rates from decay time series. *Biogeosciences*, **9**,
		3601-3612.

	[2] P.C. Hansen (1994) Regularization tools: A Matlab package for analysis
		and solution of discrete ill-posed problems. *Numerical Algorithms*, 
		**6**, 1-35.
	'''

	#extract nt and nk (or nE for daem)
	nt, nk = np.shape(model.A)

	#calculate the regularization matrix
	R = _calc_R(nk)

	#concatenate A+R and g+zeros
	A_reg = np.concatenate(
		(model.A, R*lam))

	g_reg = np.concatenate(
		(timedata.g, np.zeros(nk + 1)))

	#calculate inverse results and estimated g
	p, _ = nnls(A_reg, g_reg)
	ghat = np.inner(model.A, p)
	rgh = np.inner(R, p)

	#calculate errors
	resid = norm(timedata.g - ghat)/nt**0.5
	rgh = norm(rgh)/nk**0.5

	return p, resid, rgh

#define a function to calculate the Tikhonov regularization matrix
def _calc_R(n):
	'''
	Calculates regularization matrix (`R`) for a given size, n.

	Parameters
	----------
	n : int
		Size of the regularization matrix will be [`n+1` x `n`].

	Returns
	-------
	R : np.ndarray
		Regularization matrix of size [`n+1` x `n`].

	References
	----------
	[1] D.C. Forney and D.H. Rothman (2012) Inverse method for calculating
		respiration rates from decay time series. *Biogeosciences*, **9**,
		3601-3612.

	[2] P.C. Hansen (1994) Regularization tools: A Matlab package for analysis
		and solution of discrete ill-posed problems. *Numerical Algorithms*, 
		**6**, 1-35.
	'''

	R = np.zeros([n+1, n])

	#ensure pdf = 0 outside of E range specified
	R[0, 0] = 1.0
	R[-1, -1] = -1.0

	#1st derivative operator
	c = [-1, 1]

	#populate matrix
	for i, row in enumerate(R):
		if i != 0 and i != n:
			row[i - 1:i + 1] = c

	return R

#define function to calculte the A matrix for DAEM models
def _rpo_calc_A(E, log10omega, t, T):
	'''
	Calculates the A matrix for a DAEM model (e.g. a Ramped Pyrox run).

	Parameters
	----------
	E : array-like
		Array of activation energy points to be used in the A matrix, in kJ.
		Length `nE`.

	log10omega : scalar or array-like
		Arrhenius pre-exponential factor, either a constant value, array with
		length `nE`, or a lambda function of E.

	t : array-like
		Array of timepoints to be used in the A matrix, in seconds. Length `nt`.

	T : array-like
		Array of temperature to be used in the A matrix, in Kelvin. Length `nt`.

	Returns
	-------
	A : np.ndarray
		2d array of the Laplace transform for the Daem model. 
		Shape [`nt` x `nE`].

	References
	----------
	[1] R.L Braun and A.K. Burnham (1987) Analysis of chemical reaction 
		kinetics using a distribution of activation energies and simpler 
		models. *Energy & Fuels*, **1**, 153-161.

	[2] D.C. Forney and D.H. Rothman (2012) Inverse method for calculating
		respiration rates from decay time series. *Biogeosciences*, **9**,
		3601-3612.
	'''

	#set constants
	nt = len(t)
	nE = len(E)
	R = 8.314/1000 #kJ/mol/K

	#get arrays in the right format and ensure lengths
	E = assert_len(E, nE) #kJ
	t = assert_len(t, nt) #s
	T = assert_len(T, nt) #K

	#get log10omega into the right format
	if hasattr(log10omega,'__call__'):
		log10omega = log10omega(E)
	
	log10omega = assert_len(log10omega, nE) 
	omega = 10**log10omega #s-1

	#calculate time and E gradients
	dt = np.gradient(t)
	dE = np.gradient(E)

	#pre-allocate A
	A = np.zeros([nt,nE])

	#loop through each timestep
	for i, _ in enumerate(T):

		#generate arrays
		U = T[:i] #Kelvin, [1,i]
		dtau = dt[:i] #seconds, [1,i]

		#generate matrices
		eps_mat = np.outer(E, np.ones(i)) #kJ, [nE,i]
		omega_mat = np.outer(omega, np.ones(i)) #s-1, [nE,i]
		u_mat = np.outer(np.ones(nE), U) #Kelvin, [nE,i]

		#generate A for row i (i.e. for each timepoint) and store in A
		hE_mat = -omega_mat*dtau*np.exp(-eps_mat/(R*u_mat)) #unitless, [nE,i]
		A[i] = np.exp(np.sum(hE_mat, axis = 1))*dE #kJ/mol

	return A



'''
rampedpyrox was created as a thesis dissertation supplement by:

	Jordon D. Hemingway, MIT/WHOI Joint Program
	Currently Assistant Professor, ETH Zurich
	jordon.hemingway@erdw.ethz.ch

source code can be found at:
	
	https://github.com/FluvialSeds/rampedpyrox

documentation can be found at:

	http://rampedpyrox.readthedocs.io

Version 1.0.4 is current as of 29 May 2023 and reflects the notation used
in Hemingway et al. Biogeosciences, 2017.

To do for future versions:
* Change exception structure to "try/except"
* Add more robust nose testing suite for debugging

REVISION NOTES:
---------------

2023.05.29: CHANGED TO 'display.precison' FOR PANDAS V 2.0
2023.05.29: CHANGED 'parse_dates = True' TO PARSE AS OBJECT AND THEN
			PARSE DIRECTLY USING 'to_datetime()'. FIXED WARNING.
'''

from __future__ import(
	division,
	print_function,
	)

__version__ = '1.0.4'

__docformat__ = 'restructuredtext en'


#import timedata classes
from .timedata import(
	RpoThermogram,
	BioDecay,
	)

#import model classes
from .model import(
	Daem,
	LaplaceTransform,
	)

#import ratedata classes
from .ratedata import(
	EnergyComplex,
	kDistribution,
	)

#import results classes
from .results import(
	RpoIsotopes,
	)

#import package-level functions
from .core_functions import(
	assert_len,
	calc_L_curve,
	derivatize,
	extract_moments,
	plot_tg_isotopes,
	)



'''
This module contains helper functions for plotting rampedpyrox data.
'''


from __future__ import(
	division,
	print_function,
	)

__docformat__ = 'restructuredtext en'
__all__ = ['_bd_plot_bge',
			'_plot_dicts', 
			'_plot_dicts_iso',
			'_rem_dup_leg',
			]

import numpy as np

#define function to plot carbon flux overlaid by BGE
def _bd_plot_bge(
	t_elapsed,
	bge,
	bge_err = None,
	ax = None,
	ymin = 0.0,
	ymax = 1.0):
	'''
	Function to plot the carbon flux (in ugC min-1 L-1) overlaid by bacterial
	growth efficiency for each time bin.

	Parameters
	----------
	t_elapsed : pd.Series
		Series containing the time elapsed (in minutes), with pd.DatetimeIndex
		as index.

	bge : pd.Series
		Series containing calculated BGE values, reported at the final
		timepoint for a given value.

	bge_err : None or pd.Series
		Series containing uncertainties for BGE values, reported at the final
		timepoint for a given value. If `None`, no uncertainty is plotted.
		Defaults to `None`.

	ax : None or matplotlib.axis
		Axis to plot BGE data on. If `None`, automatically creates a
		``matplotlip.axis`` instance to return. Defaults to `None`.

	ymin : float
		Minimum y value for BGE axis. Defaults to `0.0`.

	ymax : float
		Maximum y value for BGE axis. Defaults to `1.0`.

	Returns
	-------
	ax : matplotlib.axis
		Axis containing BGE data
	'''
	#create axis if necessary and label
	if ax is None:
		_, ax = plt.subplots(1, 1)

	#find t_elapsed values for each entry in bge
	bge_inds = bge.index
	bge_times = t_elapsed[bge_inds]

	#loop through each time range and plot BGE
	for i, ind in enumerate(bge_inds[1:]):
		#find bounding time points
		t0 = t_elapsed[bge_inds[i]]
		tf = t_elapsed[ind]
		b = bge[i+1]

		#plot results
		ax.plot(
			[t0, tf],
			[b, b],
			c = 'k',
			linewidth = 2
			)

		#include uncertainty as a shaded box
		if bge_err is not None:

			berr = bge_err[i+1]

			ax.fill_between(
				[t0, tf],
				b - berr,
				b + berr,
				alpha = 0.5,
				color = 'k',
				linewidth = 0
				)

	#set limits and label
	ax.set_ylim([ymin, ymax])
	ax.set_ylabel('Bacterial Growth Efficiency (BGE)')

	return ax

#define function to pull plotting dicts
def _plot_dicts(case, td):
	'''
	Function to access different plotting dicts.

	Parameters
	----------
	case : str
		The case that defines the dict to pull.
		Acceptable strings:

			'bd_labs', \n
			'bd_md', \n
			'bd_rd', \n
			'rpo_labs', \n
			'rpo_md', \n
			'rpo_rd'
			
	td : TimeData or subclass
		``rp.TimeData`` instance containing the data to plot.

	Returns
	-------
	pl_dict : dict
		The resulting dictionary containing plotting info.
	'''

	if case == 'bd_labs':
		#create a dict to keep track of axis labels
		pl_dict = {'secs': 
						{'fraction' : ('time (s)', 'g (unitless)'),
						'rate' : ('time (s)', r'fraction/time $(s^{-1})$')
						},
					'mins' : 
						{'fraction' : ('time (min)', 'g (unitless)'),
						'rate' : ('time (min)', r'fraction/temp $(K^{-1})$')
						},
					'hours' : 
						{'fraction' : ('time (hr)', 'g (unitless)'),
						'rate' : ('time (hr)', r'fraction/temp $(K^{-1})$')
						},
					'days' : 
						{'fraction' : ('time (d)', 'g (unitless)'),
						'rate' : ('time (d)', r'fraction/temp $(K^{-1})$')}
						}

	elif case == 'bd_md':
		#create a dict to keep track of cases of modeled data
		pl_dict = {'secs': 
						{'fraction' : (td.t, td.ghat),
						'rate' : (td.t, -td.dghatdt)
						},
					'mins': 
						{'fraction' : (td.t / 60, td.ghat),
						'rate' : (td.t / 60, -td.dghatdT)
						},
					'hours': 
						{'fraction' : (td.t / (60*60), td.ghat),
						'rate' : (td.t / (60*60), -td.dghatdT)
						},
					'days': 
						{'fraction' : (td.t / (60*60*24), td.ghat),
						'rate' : (td.t / (60*60*24), -td.dghatdT)}
						}

	elif case == 'bd_rd':
		#create a dict to keep track of cases for real data
		pl_dict = {'secs': 
						{'fraction' : (td.t, td.g),
						'rate' : (td.t, -td.dgdt)
						},
					'mins': 
						{'fraction' : (td.t / 60, td.g),
						'rate' : (td.t / 60, -td.dgdT)
						},
					'hours': 
						{'fraction' : (td.t / (60*60), td.g),
						'rate' : (td.t / (60*60), -td.dgdT)
						},
					'days': 
						{'fraction' : (td.t / (60*60*24), td.g),
						'rate' : (td.t / (60*60*24), -td.dgdT)}
						}

	elif case == 'rpo_labs':
		#create a nested dict to keep track of axis labels
		pl_dict = {'time': 
						{'fraction' : ('time (s)', 'g (unitless)'),
						'rate' : ('time (s)', r'fraction/time $(s^{-1})$')
						},
					'temp' : 
						{'fraction' : ('temp (K)', 'g (unitless)'),
						'rate' : ('temp (K)', r'fraction/temp $(K^{-1})$')}
						}
	
	elif case == 'rpo_md':
		#create a nested dict to keep track of cases of modeled data
		pl_dict = {'time': 
						{'fraction' : (td.t, td.ghat),
						'rate' : (td.t, -td.dghatdt)
						},
					'temp': 
						{'fraction' : (td.T, td.ghat),
						'rate' : (td.T, -td.dghatdT)}
						}

	elif case == 'rpo_rd':
		#create a nested dict to keep track of cases for real data
		pl_dict = {'time': 
						{'fraction' : (td.t, td.g),
						'rate' : (td.t, -td.dgdt)
						},
					'temp': 
						{'fraction' : (td.T, td.g),
						'rate' : (td.T, -td.dgdT)}
						}

	return pl_dict

#define function to pull plotting dicts
def _plot_dicts_iso(case, ri):
	'''
	Function to access different plotting dicts.

	Parameters
	----------
	case : str
		The case that defines the dict to pull.
		Acceptable strings:

			'rpo_rd', \n
			'rpo_labs', \n
			'rpo_md'

	ri : Results or subclass
		``rp.Results`` instance containing the data to plot.

	Returns
	-------
	pl_dict : dict
		The resulting dictionary containing plotting info.
	'''

	if case == 'rpo_iso_labs':
		#create a nested dict to keep track of isotope result axis labels
		pl_dict = {'E': 
						{'p0E' : (r'E (kJ $mol^{-1}$)', 
							r'p(0,E)'),
						'Fm' : (r'E (kJ $mol^{-1}$)', r'Fm'),
						'd13C' : (r'E (kJ $mol^{-1}$)', 
							r'$\delta^{13}C$ (VPDB)')}
						}

	elif case == 'iso_corr':
		#create a nested dict to keep track of cases of scatter
		pl_dict = {'E': 
						{'Fm_corr' : (ri.E_frac, ri.Fm_corr, 
							ri.E_frac_std, ri.Fm_corr_std),
						'd13C_corr' : (ri.E_frac, ri.d13C_corr, 
							ri.E_frac_std, ri.d13C_corr_std)}
						}

	elif case == 'iso_raw':
		#create a nested dict to keep track of cases of scatter
		pl_dict = {'E': 
						{'Fm_raw' : (ri.E_frac, ri.Fm_raw, 
							ri.E_frac_std, ri.Fm_raw_std),
						'd13C_raw' : (ri.E_frac, ri.d13C_raw, 
							ri.E_frac_std, ri.d13C_raw_std)}
						}

	return pl_dict

#define function to remove duplicate legend entries
def _rem_dup_leg(ax):
	'''
	Removes duplicate legend entries.

	Parameters
	----------
	ax : plt.axishandle
		Axis handle containing entries to remove.

	Returns
	-------
	han_list : list
		List of axis handles.

	lab_list : list
		List of axis handle labels.
	'''
	han, lab = ax.get_legend_handles_labels()
	han_list, lab_list = [], []
	
	for h, l in zip(han, lab):
		
		if l not in lab_list:
		
			han_list.append(h)
			lab_list.append(l)

	return han_list, lab_list



'''
This package contains all the exceptions for the rampedpyrox package.
'''

#define a core exception class for subclassing
class rpException(Exception):
	'''
	Root Exception class for the rampedpyrox package. Do not call directly.
	'''
	pass

class ArrayError(rpException):
	'''
	Array-like object is not in the right form (e.g. strings).
	'''
	pass


class FileError(rpException):
	'''
	If a file does not contain the correct data.
	'''
	pass


class LengthError(rpException):
	'''
	Length of array is not what it should be.
	'''
	pass


class ScalarError(rpException):
	'''
	If something is not a scalar.
	'''
	pass


class StringError(rpException):
	'''
	If a string is not right.
	'''
	pass


'''
This module contains the TimeData superclass and all corresponding subclasses.
'''

from __future__ import(
	division,
	print_function,
	)

__docformat__ = 'restructuredtext en'
__all__ = ['RpoThermogram']

import matplotlib.pyplot as plt
import numpy as np
import warnings

from numpy.linalg import norm

#import exceptions
from .exceptions import(
	ArrayError,
	StringError,
	)

#import helper functions
from .core_functions import(
	assert_len,
	derivatize,
	)

from .plotting_helper import(
	_bd_plot_bge,
	_plot_dicts,
	_rem_dup_leg,
	)

from .summary_helper import(
	_calc_RPO_info,
	_calc_BD_info,
	)

from .model_helper import(
	_calc_ghat,
	)

from .timedata_helper import(
	_rpo_extract_tg,
	_bd_calc_bge,
	_bd_data_reduction,
	_bd_extract_profile,
	)


class TimeData(object):
	'''
	Class to store time-dependent data. Intended for subclassing, do not call
	directly.
	'''

	def __init__(self, t, T, g = None, g_std = None, T_std = None):
		'''
		Initialize the superclass.

		Parameters
		----------
		t : array-like
			Array of timepoints, in seconds. Length `nt`.

		T : array-like
			Array of temperature, in Kelvin. Length `nt`.

		g : None or array-like
			Array of the true fraction of carbon remaining at each timepoint,
			with length `nt`. Defaults to `None`.

		g_std : None or array-like
			Standard deviation of `g`, with length `nt`. Defaults to `None`.

		T_std : scalar or array-like
			The standard deviation of `T`, with length `nt`, in Kelvin. 
			Defaults to `None`.
		'''

		#store time-temperature attributes
		nt = len(t)
		self.nt = nt
		self.t = assert_len(t, nt) #s
		self.T = assert_len(T, nt) #K

		if T_std is not None:

			#only store T_std if it exists (NONE for RPO, keep for future)
			self.T_std = assert_len(T_std, nt) #K

		#store time-temperature derivatives
		self.dTdt = derivatize(self.T, self.t) #K/s

		#check if g and store
		if g is not None:

			#assert that g remains between 0 and 1
			if np.max(g) > 1 or np.min(g) < 0:
				raise ArrayError(
					'g array must remain between 0 and 1 (fractional)')

			self.g = assert_len(g, nt) #fraction

			if g_std is not None:

				#only store g_std if it exists (NONE for RPO, keep for future)
				self.g_std = assert_len(g_std, nt) #fraction

			#store g derivatives
			self.dgdt = derivatize(g, self.t)
			self.dgdT = derivatize(g, self.T)

	#define class method for creating instance directly from .csv file
	@classmethod
	def from_csv(cls, file):
		raise NotImplementedError

	#define method for forward-modeling rate data using a given model
	def forward_model(self, model, ratedata):
		'''
		Forward-models rate data for a given model and creates timedata
		instance.

		Parameters
		----------
		model : rp.Model
			``rp.Model`` instance used to calculate the forward model.

		ratedata : rp.RateData
			``rp.Ratedata instance containing the reactive continuum data.

		Warnings
		--------
		UserWarning
			If the time-temperature data in the ``rp.Model`` instance do not 
			match the time-temperature data in the ``rp.TimeData`` instance.
		'''

		#warn if self and model t and T arrays do not match
		td_type = type(self).__name__
		mod_type = type(model).__name__

		if (self.t != model.t).any() or (self.T != model.T).any():
			warnings.warn(
				'rp.TimeTata instance of type %s and rp.Model instance of'
				' type %s do not contain matching time-temperature arrays.'
				' Check that the model does not correspond to a different'
				' rp.TimeData instance' %(td_type, mod_type), UserWarning)


		#calculate forward-modelled g estimate, ghat
		ghat = _calc_ghat(model, ratedata)

		#populate with modelled data
		self.input_estimated(ghat)

	#define method for inputting the results from a model fit
	def input_estimated(self, ghat):
		'''
		Method to input modelled estimate data into ``rp.TimeData`` instance 
		and to calculate corresponding derivatives and statistics.

		Parameters
		----------
		ghat : array-like
			Array of estimated fraction of total carbon remaining at each 
			timestep. Length `nt`.
		'''

		#ensure type and size
		nt = self.nt
		ghat = assert_len(ghat, nt)

		#calculate derived attributes and store
		self.dghatdt = derivatize(ghat, self.t)
		self.dghatdT = derivatize(ghat, self.T)
		self.ghat = ghat

		#store RMSE if the model has true data, g
		if hasattr(self, 'g'):

			resid = norm(self.g - ghat)/nt**0.5
			self.resid = resid

	#define plotting method
	def plot(self, ax = None, labs = None, md = None, rd = None):
		'''
		Method for plotting ``rp.TimeData`` instance data.

		Parameters
		----------
		axis : matplotlib.axis or None
			Axis handle to plot on. Defaults to `None`.

		labs : tuple
			Tuple of axis labels, in the form (x_label, y_label).
			Defaults to `None`.

		md : tuple or None
			Tuple of modelled data, in the form  (x_data, y_data). Defaults
			to `None`.

		rd : tuple
			Tuple of real (observed) data, in the form (x_data, y_data). 
			Defaults to `None`.

		Returns
		-------
		ax : matplotlib.axis
			Updated axis handle containing data.
		'''

		#create axis if necessary and label
		if ax is None:
			_, ax = plt.subplots(1, 1)

		#label axes if labels exist
		if labs is not None:
			ax.set_xlabel(labs[0])
			ax.set_ylabel(labs[1])

		#add real data if it exists
		if rd is not None:
			ax.plot(
				rd[0], 
				rd[1],
				linewidth = 2,
				color = 'k',
				label = 'Observed Data')

			#set limits
			ax.set_xlim([0, 1.1*np.max(rd[0])])
			ax.set_ylim([0, 1.1*np.max(rd[1])])
			
		#add model-estimated data if it exists
		if md is not None:

			#plot the model-estimated total
			ax.plot(
				md[0], 
				md[1],
				linewidth = 2,
				color = 'r',
				label = 'Model-estimated data')

			#(re)set limits
			ax.set_xlim([0, 1.1*np.max(md[0])])
			ax.set_ylim([0, 1.1*np.max(md[1])])

		#remove duplicate legend entries
		han_list, lab_list = _rem_dup_leg(ax)
		
		ax.legend(
			han_list,
			lab_list, 
			loc = 'best',
			frameon = False)

		#make tight layout
		plt.tight_layout()

		return ax


class RpoThermogram(TimeData):
	__doc__='''
	Class for inputting and storing Ramped PyrOx true (observed) and estimated
	(forward-modelled) thermograms, calculating goodness of fit statistics,
	and reporting summary tables.

	Parameters
	----------
	t : array-like
		Array of time, in seconds. Length `nt`.

	T : array-like
		Array of temperature, in Kelvin. Length `nt`.

	g : None or array-like
		Array of the true fraction of carbon remaining at each timepoint,
		with length `nt`. Defaults to `None`.

	Warnings
	--------
	UserWarning
		If attempting to use isothermal data to create an ``rp.RpoThermogram``
		instance. Consider using an alternate ``rp.TimeData`` subclass (to be
		added in future versions).

	Notes
	-----
	**Important:** The inverse model used herein is highly sensitive to
	boundary effects. To avoid unnecessarily large regularizations ensure that
	inputted data are completely at baseline (ppm CO2 = 0) at the beginning
	and the end of the experiment (can use the `bl_subtract` flag to enforce
	that this is true.)

	See Also
	--------
	Daem
		``rp.Model`` subclass used to generate the distributed activation
		energy model (DAEM) transform matrix for RPO data and translate between
		time- and E-space.

	EnergyComplex
		``rp.RateData`` subclass for storing and analyzing RPO energy (rate)
		data.

	Examples
	--------
	Generating an arbitrary bare-bones thermogram containing only `t` and
	`T`::

		#import modules
		import numpy as np
		import rampedpyrox as rp

		#generate arbitrary data
		t = np.arange(1,100) #100 second experiment
		beta = 0.5 #K/second
		T = beta*t + 273.15 #K

		#create instance
		tg = rp.RpoThermogram(t,T)

	Generating a real thermogram using an RPO output .csv file and the
	``rp.RpoThermogram.from_csv`` class method, and subtracting the baseline::

		#import modules
		import rampedpyrox as rp

		#create path to data file
		file = 'path_to_folder_containing_data/thermogram_data.csv'

		#create instance using baseline-subtracted CO2 data
		tg = rp.RpoThermogram.from_csv(
			file,
			bl_subtract = True,
			nt = 250) #number of down-sampled time points

	Manually adding some model-estimated fraction remaining data as `ghat`::

		#assuming ghat has been generating using a ``rp.Daem`` model
		tg.input_estimated(ghat)

	Or, instead, you can input model-estimated g data directly from a given
	``rp.Daem`` and ``rp.EnergyComplex`` instance (*i.e.* run the forward 
	model)::

		#assuming ``rp.Daem`` named daem and ``rp.EnergyComplex`` named ec
		tg.forward_model(daem, ec)

	Plotting the resulting observed and modelled thermograms (note scatter
	when plotted against temp due to short fluctuations in measured ramp rate.
	For a "smooth" plot, always plot against time, as this is the master
	variable.)::

		#import additional modules
		import matplotlib.pyplot as plt

		#create figure
		fig, ax = plt.subplots(1,2)

		#plot resulting rates against time and temp
		ax[0] = tg.plot(
			ax = ax[0], 
			xaxis = 'time', 
			yaxis = 'rate')
		
		ax[1] = tg.plot(
			ax = ax[1], 
			xaxis = 'temp', 
			yaxis = 'rate')

	Printing a summary of the observed and modelled thermograms::

		print(tg.tg_info)
		print(tg.tghat_info)

	**Attributes**


	dghatdt : numpy.ndarray
		Array of the derivative of the estimated fraction of carbon remaining
		with respect to time at each timepoint, in fraction/second. Length 
		`nt`.

	dghatdT : numpy.ndarray
		Array of the derivative of the estimated fraction of carbon remaining
		with respect to temperature at each timepoint, in fraction/Kelvin.
		Length `nt`.

	dgdt : numpy.ndarray
		Array of the derivative of the true fraction of carbon remaining with
		respect to time at each timepoint, in fraction/second. Length `nt`.

	dgdT : numpy.ndarray
		Array of the derivative of the true fraction of carbon remaining with 
		respect to temperature at each timepoint, in fraction/Kelvin.
		Length `nt`.

	dTdt : numpy.ndarray
		Array of the derivative of temperature with respect to time (*i.e.*
		the instantaneous ramp rate) at each timepoint, in Kelvin/second.
		Length `nt`.

	g : numpy.ndarray
		Array of the true fraction of carbon remaining at each timepoint.
		Length `nt`.

	ghat : numpy.ndarray
		Array of the estimated fraction of carbon remaining at each timepoint.
		Length `nt`.

	nt : int
		Number of timepoints.

	resid : float
		The residual root mean square error (RMSE) between observed and
		modelled thermograms, g and ghat.

	t : numpy.ndarray
		Array of timepoints, in seconds. Length `nt`.

	T : numpy.ndarray
		Array of temperature, in Kelvin. Length `nt`.

	tg_info : pd.Series
		Series containing the observed thermogram summary info: 

			t_max (s), \n
			t_mean (s), \n
			t_std (s), \n
			T_max (K), \n
			T_mean (K), \n
			T_std (K), \n
			max_rate (frac/s), \n
			max_rate (frac/K), \n

	tghat_info : pd.Series
		Series containing the modelled thermogram summary info: 

			t_max (s), \n
			t_mean (s), \n
			t_std (s), \n
			T_max (K), \n
			T_mean (K), \n
			T_std (K), \n
			max_rate (frac/s), \n
			max_rate (frac/K), \n
	'''

	def __init__(self, t, T, g = None):

		#warn if T is scalar
		if isinstance(T, (int, float)) or len(set(T)) == 1:
			warnings.warn(
				'Attempting to use isothermal data for RPO run! T is a scalar'
				' value of: %r. Consider using an isothermal model type'
				' instead.' % T, UserWarning)

		super(RpoThermogram, self).__init__(
			t, 
			T, 
			g = g, 
			g_std = None, #force to be None for RPO
			T_std = None) #force to be None for RPO

		#if g exists, add RPO-specific summary file
		if g is not None:

			self.tg_info = _calc_RPO_info(self.t, self.T, self.g)

	#define class method for creating instance directly from .csv file
	@classmethod
	def from_csv(
			cls, 
			file, 
			bl_subtract = True, 
			nt = 250):
		'''
		Class method to directly import RPO data from a .csv file and create
		an ``rp.RpoThermogram`` class instance.

		Parameters
		----------
		file : str or pd.DataFrame
			File containing thermogram data, either as a path string or a
			dataframe.

		bl_subtract : Boolean
			Tells the program whether or not to linearly subtract the baseline
			such that ppmCO2 returns to 0 at the beginning and end of the run. 
			Defaults to `True`. **To minimize boundary effects, this should 
			typically be set to `True` regardless of previous data treatment.**

		nt : int
			The number of time points to use. Defaults to 250.

		Notes
		-----
		If using the `all_data` file generated by the NOSAMS RPO LabView 
		program, the date_time column must be converted to **hh:mm:ss AM/PM**
		format and a header row should be added with the following columns:

			date_time, \n
			T_room, \n
			P_room, \n
			CO2_raw, \n
			corr_int, \n
			corr_slope, \n
			temp, \n
			CO2_scaled, \n
			flow_rate, \n
			dTdt, \n
			fraction, \n
			ug_frac, \n
			ug_sum

		(Note that all columns besides `date_time`, `temp`, and `CO2_scaled`
		are unused.) Ensure that all rows before the start of temperature
		ramping and after the ovens have been turned off have been removed.

		When down-sampling, `t` contains the midpoints of each time bin and
		`g` and `T` contain the corresponding temp. and fraction remaining 
		at each midpoint.

		See Also
		--------
		RpoIsotopes.from_csv
			Classmethod for creating ``rp.RpoIsotopes`` instance directly from
			a .csv file.
		'''

		#extract data from file
		g, t, T = _rpo_extract_tg(
			file, 
			nt, 
			bl_subtract = bl_subtract)

		return cls(t, T, g = g)

	#define method for inputting forward-modelled data
	def forward_model(self, model, ratedata):
		'''
		Forward-models rate data for a given model and populates the
		thermogram with model-estimated data.

		Parameters
		----------
		model : rp.Model
			The ``rp.Daem`` instance used to calculate the forward model.

		ratedata : rp.RateData
			The ``rp.EnergyComplex`` instance containing the reactive 
			continuum data.

		Warnings
		--------
		UserWarning
			If using an an isothermal model type for an RPO run.

		UserWarning
			If using a non-energy complex ratedata type for an RPO run.

		Raises
		------
		ArrayError
			If `nE` is not the same in the ``rp.Model`` instance and the 
			``rp.RateData`` instance.

		ArrayError
			If `nt` is not the same in the ``rp.Model`` instance and the
			``rp.TimeData`` instance.

		ArrayError
			If the ``rp.RateData`` instance has no attribute `p`.

		See Also
		--------
		input_estimated
			Method used for inputting model-estimated data directly.


		EnergyComplex.inverse_model
			Class for creating an ``rp.EnergyComplex`` instance and
			calculating the inverse model.
		'''

		#warn if model is not Daem
		mod_type = type(model).__name__

		if mod_type not in ['Daem']:
			warnings.warn(
				'Attempting to calculate thermogram using a model instance of'
				' type %r. Consider using rp.Daem instance instead'
				% mod_type, UserWarning)

		#warn if ratedata is not EnergyComplex
		rd_type = type(ratedata).__name__

		if rd_type not in ['EnergyComplex']:
			warnings.warn(
				'Attempting to calculate thermogram using a ratedata instance'
				' of type %r. Consider using rp.EnergyComplex instance'
				' instead' % rd_type, UserWarning)

		#raise exception if no p data exist
		if not hasattr(ratedata, 'p'):
			raise ArrayError(
				'EnergyComplex has no p array!')

		#raise exception if not the right shape
		if model.nE != ratedata.nE:
			raise ArrayError(
				'Cannot combine model with nE = %r and RateData with'
				' nE = %r. Check that RateData was not created using'
				' a different model' % (model.nE, ratedata.nE))

		#raise exception if not the right shape
		if model.nt != self.nt:
			raise ArrayError(
				'Cannot combine model with nt = %r and TimeData with'
				' nt = %r. Check that the mode was not created using'
				' different time data' % (model.nt, self.nt))

		#call the superclass method
		super(RpoThermogram, self).forward_model(model, ratedata)

		return

	#define method for inputting model-estimate data
	def input_estimated(self, ghat):
		'''
		Inputs estimated thermogram into the ``rp.RpoThermogram`` instance and 
		calculates statistics.
		
		Parameters
		----------
		ghat : array-like
			Array of estimated fraction of total carbon remaining at each 
			timestep. Length `nt`.

		See Also
		--------
		forward_model
			Method for directly inputting estimated data from a given model
			and ratedata.	
		'''

		#call the superclass method
		super(RpoThermogram, self).input_estimated(ghat)

		#add RPO-specific modelled tg summary file
		self.tghat_info = _calc_RPO_info(self.t, self.T, ghat)


	#define plotting method
	def plot(self, ax = None, xaxis = 'time', yaxis = 'rate'):
		'''
		Plots the true and model-estimated thermograms against time or temp.

		Parameters
		----------
		ax : None or matplotlib.axis
			Axis to plot on. If `None`, automatically creates a
			``matplotlip.axis`` instance to return. Defaults to `None`.

		xaxis : str
			Sets the x axis unit, either 'time' or 'temp'. Defaults to 'time'.

		yaxis : str
			Sets the y axis unit, either 'fraction' or 'rate'. Defaults to 
			'rate'.

		Returns
		-------
		ax : matplotlib.axis
			Updated axis instance with plotted data.

		Raises
		------
		StringError
			If `xaxis` is not 'time' or 'temp'.

		StringError
			if `yaxis` is not 'fraction' or 'rate'.
		'''

		#check that axes are appropriate strings
		if xaxis not in ['time','temp']:
			raise StringError(
				'xaxis does not accept %r. Must be either "time" or "temp"'
				%xaxis)

		elif yaxis not in ['fraction','rate']:
			raise StringError(
				'yaxis does not accept %r. Must be either "rate" or'
				' "fraction"' %yaxis)

		#extract axis label ditionary
		rpo_labs = _plot_dicts('rpo_labs', self)
		labs = (
			rpo_labs[xaxis][yaxis][0], 
			rpo_labs[xaxis][yaxis][1])

		#check if real data exist
		if hasattr(self, 'g'):
			#extract real data dict
			rpo_rd = _plot_dicts('rpo_rd', self)
			rd = (
				rpo_rd[xaxis][yaxis][0], 
				rpo_rd[xaxis][yaxis][1])
		else:
			rd = None

		#check if modelled data exist
		if hasattr(self, 'ghat'):
			#extract modelled data dict
			rpo_md = _plot_dicts('rpo_md', self)
			md = (
				rpo_md[xaxis][yaxis][0], 
				rpo_md[xaxis][yaxis][1])
		else:
			md = None

		ax = super(RpoThermogram, self).plot(
			ax = ax, 
			md = md,
			labs = labs, 
			rd = rd)

		return ax


#TODO: ADD DOCUMENTATION!!

class BioDecay(TimeData):
	__doc__ = '''
	Class for inputting and storing IsoCaRB true (observed) and estimated
	(forward-modelled) decay profiles, calculating goodness of fit statistics,
	and reporting summary tables.

	Parameters
	----------
	t : array-like
		Array of time, in seconds. Length `nt`.

	T : array-like
		Array of incubation temperature, in Kelvin.

	g : None or array-like
		Array of the true fraction of carbon remaining at each timepoint,
		with length `nt`. Defaults to `None`.

	Warnings
	--------

	Notes
	-----

	See Also
	--------

	Examples
	--------

	**Attributes**
	'''

	def __init__(self, t, T, g = None):

		super(BioDecay, self).__init__(
			t, 
			T, 
			g = g, 
			g_std = None, #force to be None for BioDecay
			T_std = None) #force to be None for BioDecay

		#if g exists add Bioreactor-specific summary file 
		if g is not None:

			self.bd_info = _calc_BD_info(
				self.t, 
				self.T, 
				self.g
				)	

	#define class method for creating instance directly from .csv file
	@classmethod
	def from_csv(
		cls,
		all_file,
		sam_file,
		mins_before_zero = 30,
		Vmedia0 = 2000,
		Vhs0 = 3750,
		Fsysblk = [35, 1],
		downsampled_dt = None,
		Ctot_mano = None,
		IRGA_error = 1.0,
		mano_error = 0.01,
		bl_subtract = True,
		nt = 250,):
		'''
		Class method to directly import IsoCaRB data from a .csv file and 
		create an ``rp.BioDecay`` class instance.

		Parameters
		----------
		all_file : str or pd.DataFrame
			File containing timeseries data, either as a path string or a
			dataframe.

		sam_file : str or pd.DataFrame
			File containing sampling data (baseline checks, liquid subsampling
			times and volumes, cell counts), either as a path string or a
			dataframe. First row must contain innoculation time (t0) timestamp.

		mins_before_zero : scalar
			Number of minutes before t0 to retain in final data. Defaults to 
			`30`.

		Vmedia0 : scalar
			Initial volume of media used in experiment, in mL. Defaults to 
			`2000`.

		Vhs0 : scalar
			Initial headspace volume for experiment, in mL. Defaults to 
			`4000`.

		Fsysblk : list
			Constant flux of system blank carbon, as reported in Beaupre et 
			al., (2016), in units of ugC day-1. First value is the mean and 
			second value is the +/- 1 sigma uncertainty. Defaults to 
			`[10, 1]`.

		downsampled_dt : scalar or None:
			Timestep to be used in final, downsampled data (in minutes). If 
			`None`, no downsampling is performed and returned `all_data` will 
			have 1-minutes timesteps.

		Ctot_mano : scalar or None
			The total carbon yield as determined by the sum of manometric
			measurements for each collected fraction. Used to re-scale flux
			values to ensure sum is equal to manometric sum. If `None`, no 
			rescaling is performed. Defaults to `None`.

		IRGA_error : scalar
			One-sigma uncertainty of the infrared gas analyzer, in ppm.
			Defaults to `1`.

		mano_erro : scalar
			One-sigma relative uncertainty of the manometer, to be used when
			scaling photometric yields to manometric ones. Reported in 
			fraction of total mass (i.e. a value of 0.01 means 1% relative 
			uncertainty). Defualts to `0.01`.

		bl_subtract : Boolean
			Tells the program whether or not to linearly subtract the baseline
			such that ppmCO2 returns to 0 at the beginning and end of the run. 
			Defaults to `True`. **To minimize boundary effects, this should 
			typically be set to `True` regardless of previous data treatment.**

		nt : int
			The number of time points to use. Defaults to `250`.

		Notes
		-----
		If using the `all_data` file generated by the IsoCaRB LabView 
		program, the date_time column must be converted to 
		**m/d/yy hh:mm AM/PM** format and a header row should be added with
		the following columns:

			date_time, \n
			temp, \n
			p_room, \n
			pCO2, \n
			CO2_raw, \n
			baseline, \n
			ref, \n
			ref_meas, \n
			CO2_scaled, \n
			flow_rate, \n
			fraction, \n
			ug_frac, \n
			ug_sum

		Note that all columns besides `date_time`, `temp`, `p_room`, 
		`CO2_scaled`, and `flow_rate` are unused. Ensure that `all_file` and
		`sam_file` timestamps are aligned (i.e. this will give an error if
		`sam_file` timestamps are not contained within `all_data'.

		When down-sampling, `t` contains the midpoints of each time bin and
		`g` contains the corresponding fraction remaining at each midpoint.

		See Also
		--------
		BioIsotopes.from_csv
			Classmethod for creating ``rp.BioIsotopes`` instance directly from
			a .csv file.
		'''

		#extract all data from file and perform data reduction, blank
		# correction, and volume normalization calculations
		ad, sd = _bd_data_reduction(
			all_file,
			sam_file,
			mins_before_zero = mins_before_zero,
			Vmedia0 = Vmedia0,
			Vhs0 = Vhs0,
			Fsysblk = Fsysblk,
			downsampled_dt = downsampled_dt,
			Ctot_mano = Ctot_mano,
			IRGA_error = IRGA_error,
			mano_error = mano_error)

		#using 't_elapsed' and 'ugC_minL' columns, downsample and calculate
		# `t` and `g`, the fraction of OC remaining.

		#extract data from file (use same helper function as thermogram)
		g, t, T = _bd_extract_profile(
			ad, 
			nt, 
			bl_subtract = bl_subtract)

		#store in class instance
		bd = cls(t, T, g = g)

		#store all data and sample data information
		bd.all_data = ad
		bd.sam_data = sd

		return bd

	#define method for calculating bacterial growth efficiency
	def calc_BGE(self, cell_counts_err = None, alpha = [50, 1]):
		'''
		Function to calculate bacterial growth efficiency (BGE) over the 
		course of a biodecay experiment using 'all_data' and 'sam_data'
		files inputted using the `from_csv` method. Requires that 'sam_data'
		contains column called 'cell_ct'.

		Parameters
		----------
		cell_counts_err : None, scalar, or pd.Series
			Series containing cell count uncertainty to be used to calculate 
			BGE uncertainty. If `None`, cell counts are assumed to be known 
			perfectly. If scalar, assumed to be fractional uncertainty and 
			applied to all values (i.e. a value of 0.01 means all measurements 
			have 1 % relative uncertainty). Defaults to `None`.

		alpha : list
			List of mass of carbon per cell and associated uncertainty 
			(+/- 1 sigma), in femtograms. Defaults to `[50, 1]`.

		Raises
		------
		AttributeError
			If BioDecay object does not contain 'all_data' or 'sam_data'
			attributes.

		AttributeError
			If BioDecay object 'sam_data' table does not contain a column
			called 'cell_ct'.

		References
		----------
		PER CELL CARBON MASS REFERENCE? GET FROM NAGISSA!
		'''

		#check that object has all_data and sam_data attributes
		if not hasattr(self, 'all_data') or not hasattr(self, 'sam_data'):
			raise AttributeError(
				'bd object does not have "all_data" and/or "sam_data"' \
				' attributes. Try importing from csv files.')

		#check that sam_data has 'cell_ct' column
		if 'cell_ct' not in self.sam_data.columns:
			raise AttributeError(
				'sam_data attribute in bd object does not contain "cell_ct"' \
				' column. Double check inputted data')

		#calculate BGE
		self.BGE =  _bd_calc_bge(
			self.all_data['ugC_minL'],
			self.sam_data['cell_ct'],
			Cflux_err = self.all_data['Cflux_err'],
			cell_counts_err = cell_counts_err,
			alpha = alpha)

		return

	#define method for inputting forward-modelled data
	def forward_model(self, model, ratedata):
		'''
		Forward-models rate data for a given model and populates the
		BioDecay profile with model-estimated data.

		Parameters
		----------
		model : rp.Model
			The ``rp.LaplaceTransform`` instance used to calculate the forward
			model.

		ratedata : rp.RateData
			The ``rp.kDist`` instance containing the reactive continuum 
			data.

		Warnings
		--------
		UserWarning
			If using a non-Laplace transform model type for an IsoCaRB run.

		UserWarning
			If using a non-kDist ratedata type for an IsoCaRB run.

		Raises
		------
		ArrayError
			If `nk` is not the same in the ``rp.Model`` instance and the 
			``rp.RateData`` instance.

		ArrayError
			If `nt` is not the same in the ``rp.Model`` instance and the
			``rp.TimeData`` instance.

		ArrayError
			If the ``rp.RateData`` instance has no attribute `p`.

		See Also
		--------
		input_estimated
			Method used for inputting model-estimated data directly.


		kDist.inverse_model
			Class for creating an ``rp.kDist`` instance and
			calculating the inverse model.
		'''

		#warn if model is not LaplaceTransform
		mod_type = type(model).__name__

		if mod_type not in ['LaplaceTransform']:
			warnings.warn(
				'Attempting to calculate BioDecay using a model instance of'
				' type %r. Consider using rp.LaplaceTransform instance instead'
				% mod_type, UserWarning)

		#warn if ratedata is not kDist
		rd_type = type(ratedata).__name__

		if rd_type not in ['kDist']:
			warnings.warn(
				'Attempting to calculate BioDecay profile using a ratedata'
				' instance of type %r. Consider using rp.kDist instance'
				' instead' % rd_type, UserWarning)

		#raise exception if no p data exist
		if not hasattr(ratedata, 'p'):
			raise ArrayError(
				'kDist has no p array!')

		#raise exception if not the right shape
		if model.nE != ratedata.nE:
			raise ArrayError(
				'Cannot combine model with nk = %r and RateData with'
				' nk = %r. Check that RateData was not created using'
				' a different model' % (model.nE, ratedata.nE))

		#raise exception if not the right shape
		if model.nt != self.nt:
			raise ArrayError(
				'Cannot combine model with nt = %r and TimeData with'
				' nt = %r. Check that the mode was not created using'
				' different time data' % (model.nt, self.nt))

		#call the superclass method
		super(BioDecay, self).forward_model(model, ratedata)

		return

	#define method for inputting model-estimated data
	def input_estimated(self, ghat):
		'''
		Inputs estimated BioDecay profile into the ``rp.BioDecay`` instance and 
		calculates statistics.
		
		Parameters
		----------
		ghat : array-like
			Array of estimated fraction of total carbon remaining at each 
			timestep. Length `nt`.

		See Also
		--------
		forward_model
			Method for directly inputting estimated data from a given model
			and ratedata.	
		'''

		#call the superclass method
		super(BioDecay, self).input_estimated(ghat)

		#add BioDecay-specific modelled bd summary file
		self.bdhat_info = _calc_BD_info(self.t, self.T, ghat)

	#define method to plot timedata info (analagous to RpoThermogram.plot())
	def plot(self, ax = None, xaxis = 'secs', yaxis = 'rate'):
		'''
		Plots the true and model-estimated decay profiles against time.

		Parameters
		----------
		ax : None or matplotlib.axis
			Axis to plot on. If `None`, automatically creates a
			``matplotlip.axis`` instance to return. Defaults to `None`.

		xaxis : str
			Sets the x axis unit, either 'secs', 'mins', 'hours', or 'days'.
			Defaults to seconds.

		yaxis : str
			Sets the y axis unit, either 'fraction' or 'rate'. Defaults to 
			'rate'.

		Returns
		-------
		ax : matplotlib.axis
			Updated axis instance with plotted data.

		Raises
		------
		StringError
			if `yaxis` is not 'fraction' or 'rate'.
		'''

		#check that axes are appropriate strings
		xl = ['secs',
				'Secs',
				'seconds',
				'Seconds',
				'mins',
				'minutes',
				'Mins',
				'Minutes',
				'hours',
				'Hours',
				'days',
				'Days',
				]

		if xaxis not in xl:
			raise StringError(
				'xaxis does not accept %r. Must be "mins", "hours",'\
				'  or "days"' %xaxis)

		elif yaxis not in ['fraction','rate']:
			raise StringError(
				'yaxis does not accept %r. Must be either "rate" or'
				' "fraction"' %yaxis)

		#ensure x axis is correct for dictionary extraction
		if xaxis in ['secs','Secs','seconds','Seconds']:
			xaxis = 'secs'
		
		elif xaxis in ['mins','minuts','Mins','Minutes']:
			xaxis = 'mins'
		
		elif xaxis in ['hours','Hours']:
			xaxis = 'hours'
		
		else:
			xaxis = 'days'

		#extract axis label dictionary
		bd_labs = _plot_dicts('bd_labs', self)
		labs = (
			bd_labs[xaxis][yaxis][0], 
			bd_labs[xaxis][yaxis][1])

		#check if real data exist
		if hasattr(self, 'g'):
			#extract real data dict
			bd_rd = _plot_dicts('bd_rd', self)
			rd = (
				bd_rd[xaxis][yaxis][0], 
				bd_rd[xaxis][yaxis][1])
		else:
			rd = None

		#check if modelled data exist
		if hasattr(self, 'ghat'):
			#extract modelled data dict
			bd_md = _plot_dicts('rpo_md', self)
			md = (
				bd_md[xaxis][yaxis][0], 
				bd_md[xaxis][yaxis][1])
		else:
			md = None

		ax = super(BioDecay, self).plot(
			ax = ax, 
			md = md,
			labs = labs, 
			rd = rd)

		return ax

	#define method to plot inputted experimental results (ppmCO2, BGE, etc.)

	#TODO: MIGRATE SWITCH CASES TO PLOTTING_HELPER PLOT_DICTS FUNCTION!

	def plot_experimental(
		self,
		ax = None,
		xaxis = 'mins',
		yaxis = 'Cflux',
		overlay = None,
		overlay_ax = None):
		'''
		Plots the true and model-estimated decay profiles against time.

		Parameters
		----------

		Returns
		-------
		ax : matplotlib.axis
			Updated axis instance with plotted data.

		overlay_ax : matplotlib.axis
			Axis for overlaid plot, if it exists. Otherwise, not exported.

		Raises
		------
		AttributeError
			If BioDecay object does not contain `all_data` attribute

		AttributeError
			If BioDecay object does not contain `BGE` attribute (only
			raised if performing BGE overlay plot)

		StringError
			If `xaxis` is not 'mins', 'hours', or 'days'.

		StringError
			If `yaxis` is not 'ppmCO2' or 'Cflux'.

		StringError
			If `overlay` is not 'BGE' or None.
		'''

		#check that object has all_data and sam_data attributes
		if not hasattr(self, 'all_data'):
			raise AttributeError(
				'bd object does not have "all_data" attribute. Try importing' \
				' from csv files.')

		#check that axes are appropriate strings
		xl = ['secs',
				'Secs',
				'seconds',
				'Seconds',
				'mins',
				'minutes',
				'Mins',
				'Minutes',
				'hours',
				'Hours',
				'days',
				'Days',
				]

		if xaxis not in xl:
			raise StringError(
				'xaxis does not accept %r. Must be "mins", "hours",'\
				'  or "days"' %xaxis)

		#check that axes are appropriate strings
		yl = ['ppmCO2',
			'CO2',
			'Cflux',
			'ugC_minL',
			]

		if yaxis not in yl:
			raise StringError(
				'yaxis does not accept %r. Must be "ppmCO2" or "Cflux"' %yaxis)

		#check that overlay is the right string or none
		ol = ['bge','BGE', None]
		if overlay not in ol:
			raise StringError(
				'overlay does not accept %r. Must be "BGE" or None' %overlay)

		#check x axis input strings and extract appropriate data
		if xaxis in ['secs','Secs','seconds','Seconds']:
			x = self.all_data['t_elapsed'] * 60
			xlab = 'Elapsed time (seconds)'

		elif xaxis in ['mins','minutes','Mins','Minutes']:
			x = self.all_data['t_elapsed']
			xlab = 'Elapsed time (minutes)'

		elif xaxis in ['hours','Hours']:
			x = self.all_data['t_elapsed'] / 60
			xlab = 'Elapsed time (hours)'

		elif xaxis in ['days','Days']:
			x = self.all_data['t_elapsed'] / (60*24)
			xlab = 'Elapsed time (days)'

		#check y axis input strings and extract appropriate data
		if yaxis in ['ppmCO2','CO2']:
			y = self.all_data['CO2_nohs']
			ye = self.all_data['CO2_err']
			ylab = r'Corrected ppm $CO_{2}$'

		elif yaxis in ['Cflux','ugC_minL']:
			y = self.all_data['ugC_minL']
			ye = self.all_data['Cflux_err']
			ylab = r'Carbon flux $(\mu g C min^{-1} L^{-1})$'

		#make axis if it does not exist
		if ax is None:
			_, ax = plt.subplots(1, 1)

		#plot data
		ax.plot(x, y,
			lw = 1,
			c = [0.5, 0.5, 0.5]
			)

		ax.fill_between(x, y - ye, y + ye,
			lw = 0,
			color = [0.5, 0.5, 0.5],
			alpha = 0.3
			)

		#add labels
		ax.set_xlabel(xlab)
		ax.set_ylabel(ylab)

		#make overlay plot if necessary
		if overlay in ['bge','BGE']:

			#ensure BGE attribute exists
			if not hasattr(self, 'BGE'):
				raise AttributeError(
					'bd object does not have "BGE" attribute. Try calculating' \
					' using "calc_bge" method.')
			
			#make axis if necessary
			if overlay_ax is None:
				#twinx to overlay on same plot
				overlay_ax = ax.twinx()

			#plot BGE data
			overlay_ax =  _bd_plot_bge(
				x,
				self.BGE['mean'],
				bge_err = self.BGE['err'],
				ax = overlay_ax,
				ymin = 0.0,
				ymax = 1.0)

			#store axes
			axes = (ax, overlay_ax)

		#FUTURE ADDITIONS: Add other overplots as 'elif' statements.

		else:
			#store axis
			axes = ax

		return axes

if __name__ == '__main__':

	import rampedpyrox as rp
	


'''
This module contains helper functions for timedata classes.
'''

from __future__ import(
	division,
	print_function,
	)

__docformat__ = 'restructuredtext en'
__all__ = ['_bd_calc_bge',
			'bd_calc_telapsed',
			'_bd_calc_ugCminL',
			'_bd_correct_baseline',
			'_bd_correct_headspace',
			'_bd_data_reduction',
			'_bd_extract_profile',
			'_bd_rolling',
			'_rpo_extract_tg',
			]

import numpy as np
import pandas as pd
import scipy.signal as signal

from scipy.interpolate import interp1d

#import exceptions
from .exceptions import(
	FileError,
	)

#define function to calculate bacterial growth efficiency (BGE)
def _bd_calc_bge(
	Cflux,
	cell_counts,
	Cflux_err = None,
	cell_counts_err = None,
	alpha = [50, 1]):
	'''
	Function to calculate bacterial growth efficiency (BGE) over the course of
	a biodecay experiment.

	Parameters
	----------
	Cflux : pd.Series
		Series containing carbon flux in ugC min-1 L-1, with pd.DatetimeIndex 
		as index.

	cell_counts : pd.Series
		Series containing cell counts at each sampling point (in units of
		cells per mL), with pd.DatetimeIndex as index.

	Cflux_err : None or pd.Series
		Series containing Cflux uncertainty to be used to calculate BGE
		uncertainty. If `None`, Cflux is assumed to be known perfectly.
		Defaults to `None`.

	cell_counts_err : None, scalar, or pd.Series
		Series containing cell count uncertainty to be used to calculate BGE
		uncertainty. If `None`, cell counts are assumed to be known perfectly.
		If scalar, assumed to be fractional uncertainty and applied to all
		values (i.e. a value of 0.01 means all measurements have 1 % relative
		uncertainty). Defaults to `None`.

	alpha : list
		List of mass of carbon per cell and associated uncertainty 
		(+/- 1 sigma), in femtograms. Defaults to `[50, 1]`.

	Returns
	-------
	BGE : pd.DataFrame
		Dataframe containing calculated BGE values and associated error, 
		reported at the final timepoint for a given value.

	Raises
	------
	ValueError
		If `cell_counts` timestamps are not contained in `Cflux` index.

	References
	----------
	PER CELL CARBON MASS REFERENCE? GET FROM NAGISSA!

	'''

	#only retain timepoints with cell count values
	cell_counts.dropna(inplace = True)
	
	#calcualte indices for cell count timepoints
	cc_inds = cell_counts.index

	#check indices
	if not all([a in Cflux.index for a in cc_inds]):
		raise ValueError(
			'cell_count timestamps not contained in Cflux index.' \
			' Check timestamps!')

	#if errors are NoneType, make arrays of zeros instead
	if Cflux_err is None:
		Cflux_err = pd.Series(
			0,
			index = Cflux.index
			)

	if cell_counts_err is None:
		cell_counts_err = pd.Series(
			0,
			index = cell_counts.index
			)

	#or if scalar, broadcast out to be a Series
	elif type(cell_counts_err) in [int, float]:
		cell_counts_err = cell_counts_err * cell_counts

	#calculate microbial biomass C in ug L-1 at each time point
	cellC = alpha[0] * cell_counts * 1e-6
	cellC_err = 1e-6 * ((cell_counts * alpha[1])**2 + \
		(alpha[0] * cell_counts_err)**2 )**0.5

	#calculate difference and drop first entry (will be NaN)
	DcellC = cellC.diff()[1:]

	#calculate cellC difference uncertainty
	Dcell_counts = cell_counts.diff()[1:]
	DcellC_err = 1e-6 * ((Dcell_counts * alpha[1])**2 + \
		(alpha[0] * cell_counts_err)**2 )**0.5

	#calculate timestep in minutes
	Dt = np.gradient(Cflux.index) / pd.Timedelta(minutes = 1)

	#calculate fractions, broadcast onto Cflux index, and ffill
	nfrac = len(cell_counts)
	frac = pd.Series(np.arange(nfrac), cc_inds)
	frac = frac.reindex(
		index = Cflux.index
		)
	frac.fillna(
		method = 'bfill', 
		inplace = True
		)

	#group by fraction
	Cflux_grouped = (Cflux * Dt).groupby(frac)
	Cflux_err_grouped = ((Cflux_err * Dt)**2).groupby(frac)

	#calculate grouped sum
	DcumC_cc = Cflux_grouped.sum()
	DcumC_cc.index = cc_inds #reset index to cc_inds
	DcumC_cc[0] = np.nan #make first entry NaN (before first cell count)

	#calculate grouped error
	DcumC_cc_err = Cflux_err_grouped.sum()**0.5
	DcumC_cc_err.index = cc_inds #reset index to cc_inds
	DcumC_cc_err[0] = np.nan #make first entry NaN (before first cell count)

	#calculate fractions and associated error
	x = DcumC_cc / DcellC
	x_err = ((DcumC_cc_err / DcellC)**2 + \
		(DcumC_cc * DcellC_err / (DcellC**2) )**2)**0.5

	#calculate BGE and error
	bge = 1 / (1 + x)
	bge_err = x_err * (1 + x)**-2

	#concatenate into dataframe
	BGE = pd.concat([bge, bge_err], 
		axis = 1
		)

	#add column names
	BGE.columns = ['mean','err']

	return BGE

#define function to calculate elapsed time
def _bd_calc_telapsed(
	all_index,
	t0index):
	'''
	Function to calculate elapsed time since innoculation for biodecay
	experiment.

	Parameters
	----------
	all_index : pd.DatetimeIndex
		Index of `all_data` dataframe, as pd.DatetimeIndex. No specific 
		requirements on timestep (function calculates).

	t0index : pd.Timestamp
		Innoculation time. Must be contained in `all_data` index.

	Returns
	-------
	t_elapsed : pd.Series
		Seriers containing elapsed time since innoculation, in minutes.

	Raises
	------
	ValueError
		If t0index is not contained in all_index.
	'''

	#find where timestamp = t_0 timestamp
	try:
		ind_start = np.where(all_index == t0index)[0][0]
	
	except IndexError:
		raise ValueError(
			't0 value: %r not in all_data index. Check timestamps' % t0index)

	#calculate minutes elapsed relative to t_0 timestamp
	Dt = (all_index - all_index[ind_start])
	t_elapsed = Dt / pd.Timedelta(minutes = 1)

	return t_elapsed #in minutes

#define function to convert to carbon flux in ugC min-1 L-1
def _bd_calc_ugCminL(
	input_data,
	input_data_err,
	flow_rate,
	t_elapsed,
	p_room,
	T_room,
	Vmedia,
	Fsysblk = [10, 1],
	Ctot_mano = None,
	mano_error = 0.01):
	'''
	Function to convert ppmCO2 values into ugC min-1 L-1.

	Parameters
	----------
	input_data : pd.Series
		Series containing a ppmCO2 array to be corrected, with pd.DatetimeIndex
		as index.

	flow_rate : pd.Serires
		Series containing flow rate values (in mL/min) for the experiment,
		with pd.DatetimeIndex as index.

	t_elapsed : pd.Series
		Series containing elapsed time (in min) for the experiment, with
		pd.Datetimeindex as index.

	p_room : pd.Series
		Series containing room pressure values (in kPa) for the experiment,
		with pd.DatetimeIndex as index.

	T_room : pd.Series
		Series containing room temperature values (in C) for the experiment,
		with pd.DatetimeIndex as index.

	Vmedia : pd.Series
		Series containing volume of media (in mL) remaining at each timepoint
		in the experiment, with pd.DatetimeIndex as index.

	Fsysblk : list or None
		Constant flux of system blank carbon, as reported in Beaupre et al.,
		(2016), in units of ugC day-1. First value is the mean and second
		value is the +/- 1 sigma uncertainty. Defaults to `[10, 1]`. If `None`,
		no correction is performed.

	Ctot_mano : scalar or None
		The total carbon yield as determined by the sum of manometric
		measurements for each collected fraction. Used to re-scale flux values
		to ensure sum is equal to manometric sum. If `None`, no rescaling is
		performed. Defaults to `None`.

	mano_erro : scalar
		One-sigma relative uncertainty of the manometer, to be used when
		scaling photometric yields to manometric ones. Reported in fraction
		of total mass (i.e. a value of 0.01 means 1% relative uncertainty).
		Defualts to `0.01`.

	Returns
	-------
	Cflux : pd.Series
		Seriers containing carbon flux in ugC min-1 L-1.

	Cflux_err : pd.Series
		Series containing +/- 1sigma uncertaint of carbon flux in 
		ugC min-1 L-1.
	'''

	#calculate multiplying scalar
	R = 8.314e3 #mL*kPa/K/mol
	Mco2 = 12.01 #g/mol
	alpha = Mco2 / (R * (T_room + 273.15)) #ug/kPa/mL/ppm
	
	#calculate carbon flux and associated uncertainty
	Cflux = input_data * alpha * p_room * flow_rate
	Cflux_err = input_data_err * alpha * p_room * flow_rate

	#calculate total photometric yield
	Dt = np.gradient(input_data.index) / pd.Timedelta(minutes = 1)
	Ctot_photo = np.sum(Cflux * Dt)
	Ctot_photo_err = (np.sum((Cflux_err * Dt)**2))**0.5

	#first, re-scale to remove system blank flux
	if Fsysblk is not None:
		#calcualte total blank C
		Ctot_sysblk = Fsysblk[0] * t_elapsed[-1] / (60 * 24) #since F in days
		Ctot_sysblk_err = Fsysblk[1] * t_elapsed[-1] / (60 * 24)

		#re-scale to remove blank flux
		s1 = (Ctot_photo - Ctot_sysblk)/Ctot_photo
		s1_err = ((Ctot_sysblk_err / Ctot_photo)**2 + \
			(Ctot_sysblk * Ctot_photo_err / (Ctot_photo**2) )**2)**0.5

		#caclualted updated C flux and uncertainty
		Cflux = s1 * Cflux
		Cflux_err = ((s1_err)**2 + (Cflux_err)**2)**0.5

	#then, re-scale to match manometric yield
	if Ctot_mano is not None:
		#re-scale
		s2 = Ctot_mano / Ctot_photo

		#calculate uncertainty
		Ctot_mano_err = Ctot_mano * mano_error
		s2_err = ((Ctot_mano_err / Ctot_photo)**2 + \
			(Ctot_mano * Ctot_photo_err / (Ctot_photo**2) )**2)**0.5

		Cflux = s2 * Cflux
		Cflux_err = ((s2_err)**2 + (Cflux_err)**2)**0.5

	# finally, re-scale for volume remaining at each time point
	Cflux = 1000 * Cflux / Vmedia
	Cflux_err = 1000 * Cflux_err / Vmedia

	return Cflux, Cflux_err

#define function to correct for baseline drift
def _bd_correct_baseline(
	input_data,
	baselines,
	IRGA_error = 1):
	'''
	Function to correct biodecay ppmCO2 values for baseline drift.

	Parameters
	----------
	input_data : pd.Series
		Series containing a ppmCO2 array to be corrected, with pd.DatetimeIndex
		as index.

	baselines : pd.Series
		Series containing baseline values, with pd.DatetimeIndex as index.

	IRGA_error : scalar
		One-sigma uncertainty of the infrared gas analyzer, in ppm.
		Defaults to `1`.

	Returns
	-------
	bl_corr : pd.Series
		Seriers containing baseline corrected ppmCO2 array.

	bl_corr_err : pd.Series
		Series containing 1sigma uncertainty in baseline corrected ppmCO2
		array.

	Raises
	------
	ValueError
		If `baselines` index values are not in `input_data` index.

	Notes
	-----
	Function forces ppmCO2 at t0 to be zero. Assumes no drift after final
	baseline check.
	'''

	#extract t0 index
	t0_index = baselines.index[0]

	#only retain timepoints with baseline values
	baselines.dropna(inplace = True)
	
	#calcualte indices for baseline timepoints
	bl_inds = baselines.index

	#check indices
	if not all([a in input_data.index for a in bl_inds]):
		raise ValueError(
			'baseline timestamps not contained in all_data index.' \
			' Check timestamps!')

	#calculate indices for timepoints following bl_inds (zero timepoints)
	bl_locs = np.array([input_data.index.get_loc(a) for a in bl_inds])
	z_locs = bl_locs + 1
	z_inds = input_data.iloc[z_locs].index

	#calculate baseline CO2 array
	CO2_bl = pd.Series(index = input_data.index)
	
	#set CO2 at bl_inds to be inputted bl values, at z_inds to be zero
	CO2_bl[bl_inds] = baselines
	CO2_bl[z_inds] = 0
	
	#set baseline CO2 equal to total CO2 at intitial point and at t0
	CO2_bl[0] = input_data[0]
	CO2_bl[t0_index] = input_data[t0_index]

	#linearly interpolate
	CO2_bl.interpolate(
		method = 'linear',
		inplace = True
		)

	#subtract baseline ppmCO2 values to correct data
	bl_corr = input_data - CO2_bl

	#assume constant IRGA error for both baseline and measurement
	# total error is thus sqrt(2*(IRGA error)^2) for all time points
	err_val = (2*IRGA_error**2)**0.5
	
	bl_corr_err = pd.Series(
		err_val,
		bl_corr.index
		)

	return bl_corr, bl_corr_err

#define function to correct for headspace averaging and volume changes
def _bd_correct_headspace(
	input_data,
	flow_rate,
	samples,
	Vmedia0 = 2000,
	Vhs0 = 4000):
	'''
	Function to correct biodecay ppmCO2 values for headspace averaging and
	volume changes due to liquid sub-sampling.

	Parameters
	----------
	input_data : pd.Series
		Series containing a ppmCO2 array to be corrected, with pd.DatetimeIndex
		as index.

	flow_rate : pd.Serires
		Series containing flow rate values (in mL/min) for the experiment,
		with pd.DatetimeIndex as index.

	samples : pd.Series
		Series containing sub-sample volumnes (in mL), with pd.DatetimeIndex 
		as index.

	Vmedia0 : scalar
		Initial volume of media for experiment, in mL. Defaults to 2000.

	Vhs0 : scalar
		Initial headspace volume for experiment, in mL. Defaults to 4000.

	Returns
	-------
	hs_corr : pd.Series
		Seriers containing headspace- and volume-corrected ppmCO2 array.

	Vmedia : pd.Series
		Series of liquid media remaining at each time point.

	Raises
	------
	ValueError
		If `samples` index values are not in `input_data` index.
	'''

	#only retain timepoints where subsamples were taken
	samples.dropna(inplace = True)
	
	#calcualte indices for sampling timepoints
	sam_inds = samples.index

	#check indices
	if not all([a in input_data.index for a in sam_inds]):
		raise ValueError(
			'liq_sample timestamps not contained in all_data index.' \
			' Check timestamps!')

	#make array of volume sampled
	Vsampled = pd.Series(
		index = input_data.index
		)
	
	Vsampled[0] = 0 #initialize to zero
	Vsampled[sam_inds] = samples.cumsum()
	Vsampled = Vsampled.fillna(
		method = 'ffill'
		)

	#make arrays of volume remaining and headspace volume
	Vmedia = Vmedia0 - Vsampled
	Vhs = Vhs0 + Vsampled

	#correct ppmCO2 to acount for nonzero residence time in headspace
	# first, calculate time steps (in mins) to account for ragged arrays
	Dt = np.gradient(input_data.index) / pd.Timedelta(minutes = 1)

	# then, find first derivative of input data and filter to remove noise
	dCdt_vals = np.gradient(input_data) / Dt
	dCdt = pd.Series(
		dCdt_vals, 
		index = input_data.index)

	dCdt = _bd_rolling(dCdt, 
		window = 100,
		center = True,
		calc = 'mean')

	# finally, correct for headspace averaging using first-order, linear ODE
	hs_corr = (Vhs/flow_rate) * dCdt + input_data

	return hs_corr, Vmedia

#define function to input biodecay data and perform reduction calculations
def _bd_data_reduction(
	all_file,
	sam_file,
	mins_before_zero = 30,
	Vmedia0 = 2000,
	Vhs0 = 4000,
	Fsysblk = [10, 1],
	downsampled_dt = None,
	Ctot_mano = None,
	IRGA_error = 1.0,
	mano_error = 0.01):
	'''
	Inputs `all_data` file file from IsoCaRB instrument at Harvard and
	performs all necessary data corrections and checks.

	Parameters
	----------
	all_file : str or pd.DataFrame
		File containing timeseries data, either as a path string or a
		dataframe.

	sam_file : str or pd.DataFrame
		File containing sampling data (baseline checks, liquid subsampling
		times and volumes, cell counts), either as a path string or a
		dataframe. First row must contain innoculation time (t0) timestamp.

	mins_before_zero : scalar
		Number of minutes before t0 to retain in final data. Defaults to `30`.

	Vmedia0 : scalar
		Initial volume of media used in experiment, in mL. Defaults to `2000`.

	Vhs0 : scalar
		Initial headspace volume for experiment, in mL. Defaults to `4000`.

	Fsysblk : list
		Constant flux of system blank carbon, as reported in Beaupre et al.,
		(2016), in units of ugC day-1. First value is the mean and second
		value is the +/- 1 sigma uncertainty. Defaults to `[10, 1]`.

	downsampled_dt : scalar or None:
		Timestep to be used in final, downsampled data (in minutes). If 
		`None`, no downsampling is performed and returned `all_data` will 
		have 1-minutes timesteps.

	Ctot_mano : scalar or None
		The total carbon yield as determined by the sum of manometric
		measurements for each collected fraction. Used to re-scale flux values
		to ensure sum is equal to manometric sum. If `None`, no rescaling is
		performed. Defaults to `None`.

	IRGA_error : scalar
		One-sigma uncertainty of the infrared gas analyzer, in ppm.
		Defaults to `1`.

	mano_erro : scalar
		One-sigma relative uncertainty of the manometer, to be used when
		scaling photometric yields to manometric ones. Reported in fraction
		of total mass (i.e. a value of 0.01 means 1% relative uncertainty).
		Defualts to `0.01`.


	Returns
	-------
	all_data : pd.DataFrame
		Dataframe containing all reduced and corrected data. If inputted
		`downsampled_dt` parameter is not `None`, then `all_data` will be
		downsampled.

	sam_data : pd.DataFrame
		Dataframe containing all sampling data (baseline checks, liquid 
		subsampling times and volumes, cell counts).

	Raises
	------
	FileError
		If `all_file` is not str or ``pd.DataFrame`` instance.

	FileError
		If `sam_file` is not str or ``pd.DataFrame`` instance.

	FileError
		If index of `all_file` is not ``pd.DatetimeIndex`` instance.

	FileError
		If index of `sam_file` is not ``pd.DatetimeIndex`` instance.

	FileError
		If `all_file` does not contain "temp", "p_room", "CO2_scaled", and 
		"flow_rate" columns.

	FileError
		If `sam_file` does not contain "CO2_bl" and "liq_sample" columns.

	Notes
	-----
	This function automates all data reduction steps that were originally
	performed using Beaupre excel spreadsheets. Function was written following
	the steps in Beaupre "Steps for IsoCaRB Data Reduction" lab document.

	Error is propagated using the following assumptions:
	[1] IRGA has constant, user-defined error (typically +/-1 ppm 1sigma)
	[2] Headspace volume, gas flow rate, room temp., and room pressure are
	known exactly and have no uncertainty.
	[3] When scaling fluxes to manometric measurements, manometer has user-
	defined relative uncertainty, typically 1%.
	[4] System blank flux has constant, user-defined uncertainty (typically
	+/- 1 ugC/day 1 sigma, as defined in Beaupre et al., 2016)
	[5] Blank, baseline, and measured ppmCO2 uncertainties are independent.
	This is not strictly true, but a fine approximation.

	References
	----------
	[1] S. Beaupre et al. (2016) IsoCaRB: A novel bioreactor system to
	characterize the lability and natural carbon isotopic (14C, 13C)
	signatures of microbially respired organic matter. *L&O Methods*, **14**, 
	668-681.

	[2] N. Mahmoudi et al. (2017) Sequential bioavailability of sedimentary 
	organic matter to heterotrophic bacteria. *Environmental Microbiology*,
	**19**, 2629-2644.

	'''

	#check all_file data format and raise appropriate errors
	
	#all_file format
	if isinstance(all_file, str):
		#import as dataframe
		all_data = pd.read_csv(
			all_file,
			index_col = 0,
			parse_dates = True)

	elif isinstance(all_file, pd.DataFrame):
		#rename to all_data
		all_data = all_file

	else:
		raise FileError(
			'all_file must be pd.DataFrame instance or path string')

	#all_file data
	
	#list of necessary columns
	ad_cols = ['temp','p_room','CO2_scaled','flow_rate']

	if not isinstance(all_data.index, pd.DatetimeIndex):
		raise FileError(
			'all_file index (first column of csv file) must be' \
			' in date_time format (pd.DatetimeIndex instance)'
			)

	elif not all([a in all_data.columns for a in ad_cols]):
		raise FileError(
			'all_file must contain columns: %r' % ad_cols
			)

	#check sam_file data format and raise appropriate errors

	#sam_file format
	if isinstance(sam_file, str):
		#import as dataframe
		sam_data = pd.read_csv(
			sam_file,
			index_col = 0,
			parse_dates = True)

	elif isinstance(sam_file, pd.DataFrame):
		#rename to sam_data
		sam_data = sam_file

	else:
		raise FileError(
			'sam_file must be pd.DataFrame instance or path string')


	#sam_file data

	#list of necessary columns
	sd_cols = ['CO2_bl','liq_sample']

	if not isinstance(sam_data.index, pd.DatetimeIndex):
		raise FileError(
			'sam_file index (first column of csv file) must be' \
			' in date_time format (pd.DatetimeIndex instance)'
			)

	elif not all([a in sam_data.columns for a in sd_cols]):
		raise FileError(
			'sam_file must contain columns: %r' % ad_cols
			)

	#---------------------------------------------------#
	# 1) CALCULATE t ARRAY IN MINUTES AND REMOVE SPIKES #
	#---------------------------------------------------#

	#because seconds might have gotten dropped during import, downsample to
	# once per minute by averaging over all points in a given minute
	all_data = all_data.resample(
		'1T'
		).median()
	
	all_data.interpolate(
		method = 'linear', 
		inplace = True
		)

	#calculate time elapsed in minutes
	all_data['t_elapsed'] = _bd_calc_telapsed(
		all_data.index,
		sam_data.index[0]
		)

	#drop everything before mins_before_zero
	all_data = all_data[all_data['t_elapsed'] > -(mins_before_zero+1)]

	#calculte spike-removed CO2
	all_data['CO2_filt'] = _bd_rolling(
		all_data['CO2_scaled'],
		window = 10,
		center = True,
		calc = 'median'
		)

	#--------------------------#
	# 2) REMOVE BASELINE DRIFT #
	#--------------------------#

	all_data['CO2_blcorr'], all_data['CO2_err'] = _bd_correct_baseline(
		all_data['CO2_filt'],
		sam_data['CO2_bl'],
		IRGA_error = IRGA_error
		)

	#------------------------------------#
	# 3) CORRECT FOR HEADSPACE AVERAGING #
	#------------------------------------#

	all_data['CO2_nohs'], Vmedia = _bd_correct_headspace(
		all_data['CO2_blcorr'],
		all_data['flow_rate'],
		sam_data['liq_sample'],
		Vmedia0 = Vmedia0,
		Vhs0 = Vhs0
		)

	#-----------------------------------------------------------#
	# 4) CONVERT TO ugC/min, RESCALE, AND SUBTRACT SYSTEM BLANK #
	#-----------------------------------------------------------#

	all_data['ugC_minL'], all_data['Cflux_err'] = _bd_calc_ugCminL(
		all_data['CO2_nohs'],
		all_data['CO2_err'],
		all_data['flow_rate'],
		all_data['t_elapsed'],
		all_data['p_room'],
		all_data['temp'],
		Vmedia,
		Fsysblk = Fsysblk,
		Ctot_mano = Ctot_mano,
		mano_error = mano_error
		)

	#---------------#
	# 5) DOWNSAMPLE #
	#---------------#

	if downsampled_dt is not None:
		dt = str(downsampled_dt) + 'T'

		all_data = all_data.resample(
			dt
			).median()

	return all_data, sam_data

#define function to extract biodecay time and g arrays
def _bd_extract_profile(
	all_data, 
	nt, 
	bl_subtract = True):
	'''
	Extracts time  and carbon remaining vectors from `all_data` Dataframe
	containing reduced data as generated by the ``_bd_data_reduction`` 
	function.

	Parameters
	----------
	all_data : pd.DataFrame
		`all_data` Dataframe containing the reduced IsoCaRB carbon flux
		profile data, with pd.DatetimeIndex as index and containing a
		't_elapsed' column.

	nt : int 
		The number of time points to use.

	bl_subtract : Boolean
		Tells the program whether or not to linearly subtract the baseline
		such that ppm CO2 returns to 0 at the beginning and end of the run. 
		Treats baseline as linear from the average of the first 100 points in
		CO2 to the average of the last 100 points in CO2. Defaults to`True`.

	Returns
	-------
	g : np.ndarray
		Array of the true fraction of carbon remaining at each timepoint.
		Length `nt`.
	
	t : np.ndarray
		Array of time, in seconds. Length `nt`.

	Raises
	------
	FileError
		If `all_data` is not ``pd.DataFrame`` instance.
	
	FileError
		If index of `all_data` is not ``pd.DatetimeIndex`` instance.

	FileError
		If `all_data` does not contain "ugC_minL" and "t_elapsed" columns.

	Notes
	-----
	Noisy data, especially at the beginning of the run, could lead to `g`
	values slightly outside of the (0, 1) range (*i.e.* noisy ppm CO2 less than
	zero could lead to slightly negative `g`). This method removes this
	possibility by enforcing that all values of `g` are within (0, 1).

	'''

	#check data format and raise appropriate errors
	if not isinstance(all_data, pd.DataFrame):
		raise FileError(
			'all_data must be pd.DataFrame instance')

	if 'ugC_minL' and 't_elapsed' and 'temp' not in all_data.columns:
		raise FileError(
			'all_data must have "ugC_minL", "t_elapsed", and "temp" columns')

	#drop data before t = 0
	ad = all_data[all_data['t_elapsed'] >= 0]

	#extract necessary data
	secs_m = ad['t_elapsed'] * 60 #conver to s
	Cflux_m = ad['ugC_minL']
	T_m = ad['temp']

	#before continuing, profile must be linearly interpolated to deal
	# with missing data (e.g. if data deleted during data clean-up).

	#make array of 1-minute deltas
	secs = np.arange(0,np.max(secs_m), 60)

	#make CO2 function and interpolate
	fsm = interp1d(secs_m, Cflux_m)
	Cflux = fsm(secs)

	#make Temp function and interpolate
	ftm = interp1d(secs_m, T_m)
	Temp = ftm(secs)

	#linearly subtract baseline if required
	if bl_subtract is True:
		#calculate initial and final bl
		#TODO: POTENTIALLY UPDATE THIS IF REAL SAMPLES HAVE LONGER INITIAL DELAY BEFORE RISING!!
		bl0 = Cflux[0] 
		blf = np.average(Cflux[-100:])

		#subtract linear baseline
		npt = len(Cflux)
		bl = bl0 + (blf - bl0)*np.arange(0, npt)/npt

		Cflux = Cflux - bl

	#calculate total CO2
	tot = np.sum(Cflux)

	#calculate alpha
	alpha = np.cumsum(Cflux)/tot

	#assert that alpha remains between 0 and 1 (noisy data at the beginning 
	# and end of the run could lead to alpha *slightly* outside fo this range)
	alpha[alpha > 1.0] = 1.0
	alpha[alpha < 0.0] = 0.0

	#generate t array
	t0 = secs[0]; tf = secs[-1]
	dt = (tf-t0)/nt

	#make downsampled points at midpoint
	t = np.linspace(t0, tf, nt + 1) + dt/2 
	
	#drop last point since it's beyond tf
	t = t[:-1] 

	#generate functions to down-sample
	fT = interp1d(secs, Temp)
	fg = interp1d(secs, alpha)
	
	#create downsampled arrays
	T = fT(t) + 273.15 #convert to K
	g = 1-fg(t)

	return g, t, T

#define function to calculate rolling values
def _bd_rolling(
	input_data,
	window = 10,
	center = True,
	calc = 'median'):
	'''
	Function to calculate rolling values (means, medians, etc.) for biodecay
	experiment.

	Parameters
	----------
	input_data : pd.Series
		Series containing a ppmCO2 array to be rolled.

	window : int
		Number of timesteps to integrate over for rolling values. Devaults to
		10.

	center : boolean
		Tells the funciton whether or not to center the rolling values.
		Defaults to `True`.

	calc : str
		Type of rolling value to take. Either 'mean', 'median', or 'sum'.
		Defaults to 'median'.

	Returns
	-------
	rolled : pd.Series
		Seriers containing rolled ppmCO2 array.

	Raises
	------
	ValueError
		If `calc` is not one of 'mean', 'median', or 'sum'.

	Notes
	-----
	`window` is the window width *not* the time width. If `input_data`
	contains uneven timesteps, then `window` will not be a constant size in
	time. (termed "ragged" in pandas notation.)
	'''

	#raise errors
	if calc not in ['mean','median','sum']:
		raise ValueError(
			'calc: %r not recognized. Must be mean, median, or sum' % calc)

	#calculate rolling
	roll = input_data.rolling(
		window = window,
		center = center,
		)

	#execute rolling
	rolled = getattr(roll, calc)()

	#forward- and back-fill nans
	rolled = rolled.fillna(
		method = 'ffill'
		)

	rolled = rolled.fillna(
		method = 'bfill'
		)

	return rolled

#define function to extract variables from .csv file
def _rpo_extract_tg(
	file, 
	nt, 
	bl_subtract = True):
	'''
	Extracts time, temperature, and carbon remaining vectors from `all_data`
	file generated by NOSAMS RPO LabView program.

	Parameters
	----------
	file : str or pd.DataFrame
		File containing thermogram data, either as a path string or a
		dataframe.

	nt : int 
		The number of time points to use.

	bl_subtract : Boolean
		Tells the program whether or not to linearly subtract the baseline
		such that ppm CO2 returns to 0 at the beginning and end of the run. 
		Treats baseline as linear from the average of the first 100 points in
		CO2 to the average of the last 100 points in CO2. Defaults to`True`.

	Returns
	-------
	g : np.ndarray
		Array of the true fraction of carbon remaining at each timepoint.
		Length `nt`.
	
	t : np.ndarray
		Array of time, in seconds. Length `nt`.

	T : np.ndarray
		Array of temperature, in Kelvin. Length `nt`.

	Raises
	------
	FileError
		If `file` is not str or ``pd.DataFrame`` instance.
	
	FileError
		If index of `file` is not ``pd.DatetimeIndex`` instance.

	FileError
		If `file` does not contain "CO2_scaled" and "temp" columns.

	Notes
	-----
	Noisy data, especially at the beginning of the run, could lead to `g`
	values slightly outside of the (0, 1) range (*i.e.* noisy ppm CO2 less than
	zero could lead to slightly negative `g`). This method removes this
	possibility by enforcing that all values of `g` are within (0, 1).

	'''

	#check data format and raise appropriate errors
	if isinstance(file, str):

# REVISION NOTES
# 2023.05.29: CHANGED 'parse_dates = True' TO PARSE AS OBJECT AND THEN
#			  PARSE DIRECTLY USING 'to_datetime()'. FIXED WARNING.

		#import as dataframe
		file = pd.read_csv(
			file,
			index_col = 0,
			# parse_dates = True
			parse_dates = False
			)

		#convert index to datetime
		ind = file.index.astype(object)
		ind_dt = pd.to_datetime(ind,format='%I:%M:%S %p')

		#now re-save as index
		file.index = ind_dt

	elif not isinstance(file, pd.DataFrame):
		raise FileError(
			'file must be pd.DataFrame instance or path string')

	if 'CO2_scaled' and 'temp' not in file.columns:
		raise FileError(
			'file must have "CO2_scaled" and "temp" columns')

	elif not isinstance(file.index, pd.DatetimeIndex):
		raise FileError(
			'file index must be pd.DatetimeIndex instance')

	#extract necessary data
	secs_m = (file.index - file.index[0]).seconds.values
	CO2_m = file.CO2_scaled
	T_m = file.temp

	#before continuing, thermogram must be linearly interpolated to deal
	# with missing data (e.g. if data deleted during data clean-up).
	# Version: 0.1.3., bug noticed by Cristina Subt (SFU).

	#make array of 1-second deltas
	secs = np.arange(0,np.max(secs_m))

	#make CO2 function and interpolate
	fsm = interp1d(secs_m, CO2_m)
	CO2 = fsm(secs)

	#make Temp function and interpolate
	ftm = interp1d(secs_m, T_m)
	Temp = ftm(secs)

	#stop Version 0.1.3. bug fix here.

	#linearly subtract baseline if required
	if bl_subtract is True:
		#calculate initial and final bl
		bl0 = np.average(CO2[:100])
		blf = np.average(CO2[-100:])

		#subtract linear baseline
		npt = len(CO2)
		bl = bl0 + (blf - bl0)*np.arange(0, npt)/npt

		CO2 = CO2 - bl

	#calculate total CO2
	tot = np.sum(CO2)

	#calculate alpha
	alpha = np.cumsum(CO2)/tot

	#assert that alpha remains between 0 and 1 (noisy data at the beginning 
	# and end of the run could lead to alpha *slightly* outside fo this range)
	alpha[alpha > 1.0] = 1.0
	alpha[alpha < 0.0] = 0.0

	#generate t array
	t0 = secs[0]; tf = secs[-1]
	dt = (tf-t0)/nt

	#make downsampled points at midpoint
	t = np.linspace(t0, tf, nt + 1) + dt/2 
	
	#drop last point since it's beyond tf
	t = t[:-1] 

	#generate functions to down-sample
	fT = interp1d(secs, Temp)
	fg = interp1d(secs, alpha)
	
	#create downsampled arrays
	T = fT(t) + 273.15 #convert to K
	g = 1-fg(t)

	return g, t, T



'''
This module contains the RateData superclass and all corresponding subclasses.
'''

from __future__ import(
	division,
	print_function,
	)

__docformat__ = 'restructuredtext en'
__all__ = ['EnergyComplex']

#import modules
import matplotlib.pyplot as plt
import numpy as np
import warnings

from numpy.linalg import norm

#import exceptions
from .exceptions import(
	ArrayError,
	ScalarError,
	)

#import helper functions
from .core_functions import(
	assert_len,
	)

from .plotting_helper import(
	_rem_dup_leg,
	)

from .summary_helper import(
	_calc_rate_info,
	)

from .model_helper import(
	_calc_p,
	)

class RateData(object):
	'''
	Class to store rate-dependent data. Intended for subclassing, do not call
	directly.
	'''

	def __init__(self):
		'''
		Initialize the superclass

		Parameters
		----------
		k : array-like
			Array of k/E values considered in the model. Length `nk`.

		p : array-like
			Array of a pdf of the distribution of k/E values. Length `nk`.
		'''
		raise NotImplementedError

	#define classmethod to generate instance by inverse modeling timedata with
	# a given model
	@classmethod
	def inverse_model(
			cls, 
			model, 
			timedata, 
			lam = 'auto'):
		'''
		Inverse models an ``rp.TimeData`` instance using a given ``rp.Model``
		instance and creates an ``rp.RateData`` instance.

		Parameters
		----------
		model : rp.Model
			``rp.Model`` instance containing the A matrix to use for inversion.

		timedata : rp.TimeData
			``rp.TimeData`` instance containing the timeseries data to invert.

		Raises
		------
		ScalarError
			If `lam` is not scalar or 'auto'.

		Warnings
		--------
		UserWarning
			If ``scipy.optimize.least_squares`` cannot converge on a 
			solution.

		See Also
		--------
		TimeData.forward_model
			``rp.TimeData`` method for forward-modeling an ``rp.RateData`` instance
			using a particular model.
		'''

		#extract model rate/E and store as k variable (necessary since models
		# have different nomenclature)
		if hasattr(model, 'k'):
			k = model.k
		
		elif hasattr(model, 'E'):
			k = model.E

		#calculate best-fit lambda value if necessary
		if lam in ['auto', 'Auto']:
			lam = model.calc_L_curve(timedata, plot = False)
		
		elif isinstance(lam, (int, float)):
			lam = float(lam)
		
		else:
			raise ScalarError(
				'lam must be int, float, or "auto"')

		#generate regularized pdf, p
		p, resid, rgh = _calc_p(model, timedata, lam)

		#create class instance
		rd = cls(k, p = p)

		#input estimated data
		rd.input_estimated(
			lam = lam,
			resid = resid,
			rgh = rgh)

		return rd

	#define a method to input estimated rate data
	def input_estimated(
			self,
			lam = None, 
			resid = None, 
			rgh = None):
		'''
		Inputs estimated data into an ``rp.RateData`` instance.

		Parameters
		----------		
		lam : scalar
			Best-fit smoothing weighting factor for Tikhonov regularization.
			Calculated using L-curve approach.

		resid : float
			Residual RMSE from inverse model.

		rgh : float
			Roughness from inverse model.

		Raises
		------
		ScalarError
			If lam is not scalar or `None`.
		'''

		#extract n rate/E (necessary since models have different nomenclature)
		if hasattr(self, 'nk'):
			nk = self.nk
			k = self.k
		
		elif hasattr(self, 'nE'):
			nk = self.nE
			k = self.E

		#store attributes
		self.resid = resid
		self.rgh = rgh

		#input lam if it exists for bookkeeping
		if lam is not None:
			if not isinstance(lam, (int, float)):
				raise ScalarError(
					'lam must be None, int, or float')
			
			else:
				self.lam = lam

	#define plotting method
	def plot(self, ax = None, labs = None, rd = None):
		'''
		Method for plotting ``rp.RateData`` instance data.

		Parameters
		----------
		axis : matplotlib.axis or None
			Axis handle to plot on. Defaults to `None`.

		labs : tuple
			Tuple of axis labels, in the form (x_label, y_label).
			Defaults to `None`.

		rd : tuple
			Tuple of real data, in the form (x_data, y_data).

		Returns
		-------
		ax : matplotlib.axis
			Updated axis handle containing data.
		'''

		#create axis if necessary and label
		if ax is None:
			_, ax = plt.subplots(1, 1)

		#label axes
		if labs is not None:
			ax.set_xlabel(labs[0])
			ax.set_ylabel(labs[1])

		#add real data if it exists
		if rd is not None:
			#plot real data
			ax.plot(
				rd[0], 
				rd[1],
				linewidth = 2,
				color = 'k',
				label = r'Regularized p(0,E) ($\lambda$ = %.2f)' %self.lam)

			#set limits
			# ax.set_xlim([0, 1.1*np.max(rd[0])])
			# ax.set_ylim([0, 1.1*np.max(rd[1])])

		#remove duplicate legend entries
		han_list, lab_list = _rem_dup_leg(ax)
		
		ax.legend(
			han_list,
			lab_list, 
			loc = 'best',
			frameon = False)

		#make tight layout
		plt.tight_layout()

		return ax


class EnergyComplex(RateData):
	__doc__='''
	Class for inputting and storing Ramped PryOx activation energy 
	distributions.

	Parameters
	----------
	E : array-like
		Array of activation energy, in kJ/mol. Length `nE`.

	p : None or array-like
		Array of the regularized pdf of the E distribution, p(0,E). Length
		`nE`. Defaults to `None`.

	Raises
	------
	ArrayError
		If the any value in `E` is negative.

	See Also
	--------
	Daem
		``rp.Model`` subclass used to generate the Laplace transform for RPO
		data and translate between time- and E-space.

	RpoThermogram
		``rp.TimeData`` subclass containing the time and fraction remaining data
		used for the inversion.

	Examples
	--------
	Generating a bare-bones energy complex containing only `E` and `p`::

		#import modules
		import rampedpyrox as rp
		import numpy as np

		#generate arbitrary Gaussian data
		E = np.arange(50, 350)

		def Gaussian(x, mu, sig):
			scalar = (1/np.sqrt(2*np.pi*sig**2))*
			y = scalar*np.exp(-(x-mu)**2/(2*sig**2))
			return y

		p = Gaussian(E, 150, 10)

		#create the instance
		ec = rp.EnergyComplex(E, p = p)

	Or, insteand run the inversion to generate an energy complex using an 
	``rp.RpoThermogram`` instance, tg, and an ``rp.Daem`` instance, daem::

		#keeping defaults, not combining any peaks
		ec = rp.EnergyComplex(
			daem, 
			tg, 
			lam = 'auto')

	Plotting the resulting regularized energy complex::

		#import additional modules
		import matplotlib.pyplot as plt

		#create figure
		fig, ax = plt.subplots(1,1)

		#plot resulting E pdf, p(0,E)
		ax = ec.plot(ax = ax)

	**Attributes**

	E : np.ndarray
		Array of activation energy, in kJ/mol. Length `nE`.

	nE : int
		Number of E points.

	ec_info : pd.Series
		Series containing the observed EnergyComplex summary info: 

			E_max (kJ/mol), \n
			E_mean (kJ/mol), \n
			E_std (kJ/mol), \n
			p0E_max

	lam : float
		Tikhonov regularization weighting factor.

	p : np.ndarray
		Array of the pdf of the E distribution, p0E. Length `nE`.

	resid : float
		The RMSE between the measured thermogram data and the estimated 
		thermogram using the p (ghat). Used for determining the best-fit lambda
		value.

	rgh :
		The roughness RMSE. Used for determining best-fit lambda value.

	References
	----------
	[1] B. Cramer (2004) Methane generation from coal during open system 
		pyrolysis investigated by isotope specific, Gaussian distributed 
		reaction kinetics. *Organic Geochemistry*, **35**, 379-392.

	[2] D.C. Forney and D.H. Rothman (2012) Common structure in the 
		heterogeneity of plant-matter decay. *Journal of the Royal Society*
		*Interface*, rsif.2012.0122.

	[3] D.C. Forney and D.H. Rothman (2012) Inverse method for calculating
		respiration rates from decay time series. *Biogeosciences*, **9**,
		3601-3612.
	'''

	def __init__(self, E, p = None):

		#store activation energy attributes
		nE = len(E)

		#ensure types
		E = assert_len(E, nE)

		#ensure that E is non-negative
		if np.min(E) < 0:
			raise ArrayError(
				'Minimum value for E is: %r. Elements in E must be'
				' non-negative.' % np.min(E))

		self.E = E
		self.nE = nE

		#check if p exists and store p, statistics
		if p is not None:
			self.p = assert_len(p, nE)
			self.ec_info = _calc_rate_info(E, p, kstr = 'E')

	#define classmethod to generate instance by inverse modeling timedata with
	# a model
	@classmethod
	def inverse_model(
			cls, 
			model, 
			timedata, 
			lam = 'auto'):
		'''
		Generates an energy complex by inverting an ``rp.TimeData`` instance 
		using a given ``rp.Model`` instance.

		Parameters
		----------
		model : rp.Model
			``rp.Model`` instance containing the A matrix to use for 
			inversion.

		timedata : rp.TimeData
			``rp.TimeData`` instance containing the timeseries data to invert.

		lam : scalar or 'auto'
			Smoothing weighting factor for Tikhonov regularization. Defaults
			to 'auto'.

		Warnings
		--------
		UserWarning
			If ``scipy.optimize.least_squares`` cannot converge on a solution.

		UserWarning
			If attempting to use timedata that is not a ``rp.RpoThermogram``
			instance.

		UserWarning
			If attempting to use a model that is not a ``rp.Daem`` instance.

		See Also
		--------
		RpoThermogram.forward_model
			``rp.TimeData`` method for forward-modeling an ``rp.RateData`` 
			instance using a particular model.
		'''

		#warn if model is not Daem
		mod_type = type(model).__name__

		if mod_type not in ['Daem']:
			warnings.warn(
				'Attempting to calculate isotopes using a model instance of'
				' type %r. Consider using rp.Daem instance instead'
				% mod_type, UserWarning)

		#warn if timedata is not RpoThermogram
		td_type = type(timedata).__name__

		if td_type not in ['RpoThermogram']:
			warnings.warn(
				'Attempting to calculate isotopes using an isothermal timedata'
				' instance of type %r. Consider using rp.RpoThermogram' 
				' instance instead' % td_type, UserWarning)

		ec = super(EnergyComplex, cls).inverse_model(
			model, 
			timedata,
			lam = lam)

		return ec

	#define a method to input estimated rate data
	def input_estimated(
			self, 
			lam = 0, 
			resid = 0, 
			rgh = 0):
		'''
		Inputs estimated rate data into the ``rp.EnergyComplex`` instance and
		calculates statistics.

		Parameters
		----------
		lam : scalar
			Tikhonov regularization weighting factor used to generate
			estimated data. Defaults to 0.

		resid : float
			Residual RMSE for the inputted estimated data. Defaults to 0.

		rgh : float
			Roughness RMSE for the inputted estimated data. Defaults to 0.
		'''

		super(EnergyComplex, self).input_estimated(
			lam = lam,
			resid = resid,
			rgh = rgh)

	#define plotting method
	def plot(self, ax = None):
		'''
		Plots the pdf of E, p(0,E), against E.

		Keyword Arguments
		-----------------
		ax : None or matplotlib.axis
			Axis to plot on. If `None`, automatically creates a
			``matplotlip.axis`` instance to return. Defaults to `None`.

		Returns
		-------
		ax : matplotlib.axis
			Updated axis instance with plotted data.
		'''

		#create axis label tuple
		labs = (r'E (kJ/mol)', r'$p(0, E)$')

		#check if data exist
		if hasattr(self, 'p'):
			#extract data
			rd = (self.E, self.p)
		else:
			rd = None

		ax = super(EnergyComplex, self).plot(
			ax = ax, 
			labs = labs, 
			rd = rd)

		return ax


class kDistribution(RateData):
	__doc__='''

	ADD DOCSTRING!

	'''

	def __init__(self, k, p = None):

		#store activation energy attributes
		nk = len(k)

		#ensure types
		k = assert_len(k, nk)

		#store
		self.k = k
		self.nk = nk

		#check if p exists and store p, statistics
		if p is not None:
			self.p = assert_len(p, nk)
			self.kd_info = _calc_rate_info(k, p, kstr = 'k')

	#define classmethod to generate instance by inverse modeling timedata with
	# a model
	@classmethod
	def inverse_model(
			cls, 
			model, 
			timedata, 
			lam = 'auto'):
		'''
		Generates an energy complex by inverting an ``rp.TimeData`` instance 
		using a given ``rp.Model`` instance.

		Parameters
		----------
		model : rp.Model
			``rp.Model`` instance containing the A matrix to use for 
			inversion.

		timedata : rp.TimeData
			``rp.TimeData`` instance containing the timeseries data to invert.

		lam : scalar or 'auto'
			Smoothing weighting factor for Tikhonov regularization. Defaults
			to 'auto'.

		Warnings
		--------
		UserWarning
			If ``scipy.optimize.least_squares`` cannot converge on a solution.

		UserWarning
			If attempting to use timedata that is not a ``rp.BioDecay``
			instance.

		UserWarning
			If attempting to use a model that is not a ``rp.LaplaceTransform``
			instance.

		See Also
		--------
		BioDecay.forward_model
			``rp.TimeData`` method for forward-modeling an ``rp.RateData`` 
			instance using a particular model.
		'''

		#warn if model is not LaplaceTransform
		mod_type = type(model).__name__

		if mod_type not in ['LaplaceTransform']:
			warnings.warn(
				'Attempting to calculate p distribution using a model instance'
				' of type %r. Consider using rp.LaplaceTransform instance'
				' instead' % mod_type, UserWarning)

		#warn if timedata is not RpoThermogram
		td_type = type(timedata).__name__

		if td_type not in ['BioDecay']:
			warnings.warn(
				'Attempting to calculate p distribution using a non-isothermal'
				' timedata instance of type %r. Consider using rp.BioDecay' 
				' instance instead' % td_type, UserWarning)

		ec = super(kDistribution, cls).inverse_model(
			model, 
			timedata,
			lam = lam)

		return ec

	#define a method to input estimated rate data
	def input_estimated(
			self, 
			lam = 0, 
			resid = 0, 
			rgh = 0):
		'''
		Inputs estimated rate data into the ``rp.kDistribution`` instance and
		calculates statistics.

		Parameters
		----------
		lam : scalar
			Tikhonov regularization weighting factor used to generate
			estimated data. Defaults to 0.

		resid : float
			Residual RMSE for the inputted estimated data. Defaults to 0.

		rgh : float
			Roughness RMSE for the inputted estimated data. Defaults to 0.
		'''

		super(kDistribution, self).input_estimated(
			lam = lam,
			resid = resid,
			rgh = rgh)


	#define plotting method
	def plot(self, ax = None):
		'''
		Plots the pdf of k, p(0,k), against k.

		Keyword Arguments
		-----------------
		ax : None or matplotlib.axis
			Axis to plot on. If `None`, automatically creates a
			``matplotlip.axis`` instance to return. Defaults to `None`.

		Returns
		-------
		ax : matplotlib.axis
			Updated axis instance with plotted data.
		'''

		#create axis label tuple
		labs = (r'k $(s^{-1})$', r'$p(0, k)$')

		#check if data exist
		if hasattr(self, 'p'):
			#extract data
			rd = (self.k, self.p)
		else:
			rd = None

		ax = super(kDistribution, self).plot(
			ax = ax, 
			labs = labs, 
			rd = rd)

		return ax


if __name__ == '__main__':

	import rampedpyrox as rp



'''
This module contains helper functions for generating rampedpyrox summary tables.
'''

from __future__ import(
	division,
	print_function,
	)

__docformat__ = 'restructuredtext en'
__all__ = [
	'_calc_rate_info', 
	'_calc_BD_info', 
	'_calc_RPO_info', 
	'_calc_ri_info'
	]

import numpy as np
import pandas as pd

#import helper functions
from .core_functions import(
	derivatize,
	extract_moments
	)

#define function to calculate ratedata info and store
def _calc_rate_info(k, p, kstr = 'E'):
	'''
	Calculates the ``rp.RateData`` instance summary statistics and stores in
	a Series.

	Parameters
	----------
	k : np.ndarray
		Array of rates (or E), length `nk`.
	p : np.ndarray
		Array of pdf of rates (or E), length `nk`.
	kstr : string
		String of nomenclature for k (i.e. 'k' or 'E')

	Returns
	-------
	rate_summary : pd.Series
		Series of resulting RateData summary info.
	'''

	# REVISION NOTES:
	#	2023.05.29: CHANGED TO 'display.precison' FOR PANDAS V 2.0
	
	#set pandas display options
	pd.set_option('display.precision', 2)

	if kstr == 'E':
		unit = ' (kJ/mol)'

	else:
		unit = ' (s-1)'

	#define series index
	ind = [kstr + '_max' + unit,
			kstr + '_mean' + unit,
			kstr + '_std' + unit,
			'p0(' + kstr + ')_max']

	#find max
	i = np.where(p == np.max(p))[0][0]

	#calculate statistics
	kmax = k[i]
	pmax = p[i]
	kav, kstd = extract_moments(k, p)

	#combine into list
	vals = [kmax, kav, kstd, pmax]

	#make series
	rate_summary = pd.Series(vals, index = ind)

	return rate_summary

#define function to calculate BioDecay timedata info and store
def _calc_BD_info(t, T, g):
	'''
	Calculates the ``rp.TimeData`` instance BioDecay summary statistics and
	stores in a Series.

	Parameters
	----------
	t : numpy.ndarray
		Array of timepoints, in seconds. Length `nt`.
	T : numpy.ndarray
		Array of temperature, in Kelvin. Length `nt`.
	g : numpy.ndarray
		Array of the true fraction of carbon remaining at each timepoint.
		Length `nt`.

	Returns
	-------
	bd_summary : pd.Series
		Series of resulting BioDecay summary info.
	'''

	# REVISION NOTES:
	#	2023.05.29: CHANGED TO 'display.precison' FOR PANDAS V 2.0
	
	#set pandas display options
	pd.set_option('display.precision', 2)

	#define series index
	ind = ['t_max (s)',
			't_mean (s)',
			't_std (s)',
			'max_rate (frac/s)']

	#derivatize g
	dgdt = derivatize(g,t)

	#find max
	i = np.where(dgdt == np.min(dgdt))[0][0]

	#calculate statistics
	tmax = t[i]
	rmax_s = -dgdt[i]
	tav, tstd = extract_moments(t, dgdt)

	#combine into list
	vals = [tmax, tav, tstd, rmax_s]

	#make series
	bd_summary = pd.Series(vals, index = ind)

	return bd_summary

#define function to calculate RampedPyrox timedata info and store
def _calc_RPO_info(t, T, g):
	'''
	Calculates the ``rp.TimeData`` instance thermogram summary statistics and
	stores in a Series.

	Parameters
	----------
	t : numpy.ndarray
		Array of timepoints, in seconds. Length `nt`.
	T : numpy.ndarray
		Array of temperature, in Kelvin. Length `nt`.
	g : numpy.ndarray
		Array of the true fraction of carbon remaining at each timepoint.
		Length `nt`.

	Returns
	-------
	tg_summary : pd.Series
		Series of resulting thermogram summary info.
	'''

	# REVISION NOTES:
	#	2023.05.29: CHANGED TO 'display.precison' FOR PANDAS V 2.0
	
	#set pandas display options
	pd.set_option('display.precision', 2)

	#define series index
	ind = ['t_max (s)',
			't_mean (s)',
			't_std (s)',
			'T_max (K)',
			'T_mean (K)',
			'T_std (K)',
			'max_rate (frac/s)',
			'max_rate (frac/K)']

	#derivatize g
	dgdt = derivatize(g,t)
	dgdT = derivatize(g,T)

	#find max
	i = np.where(dgdt == np.min(dgdt))[0][0]

	#calculate statistics
	tmax = t[i]
	Tmax = T[i]
	rmax_s = -dgdt[i]
	rmax_k = -dgdT[i]
	tav, tstd = extract_moments(t, dgdt)
	Tav, Tstd = extract_moments(T, dgdT)

	#combine into list
	vals = [tmax, tav, tstd, Tmax, Tav, Tstd, rmax_s, rmax_k]

	#make series
	tg_summary = pd.Series(vals, index = ind)

	return tg_summary

#create method for calculating RPO isotope summary info table
def _calc_ri_info(ri, flag = 'raw'):
	'''
	Calculates the ``rp.RpoIsotopes`` instance fraction info and stores as a
	DataFrame.

	Parameters
	----------
	ri : rp.RpoIsotopes
		``rp.RpoIsotopes`` instance containing fractions to be summarized.
	flag : str
		Tells the method whether to store raw or corrected data
	
	Returns
	-------
	frac_info : pd.DataFrame
		DataFrame of resulting fraction info.
	'''

	#create empty list to store existing data
	info = []

	#create empty list to store name strings
	names = []

	#add t_frac
	info.append(ri.t_frac[:,0])
	info.append(ri.t_frac[:,1])
	names.append('t0 (s)')
	names.append('tf (s)')

	#add E
	info.append(ri.E_frac)
	info.append(ri.E_frac_std)
	names.append('E (kJ/mol)')
	names.append('E std. (kJ/mol)')

	#go through each measurement and add if it exists
	m = 'm_' + flag
	m_std = 'm_' + flag + '_std'

	if hasattr(ri, m):
		
		info.append(getattr(ri, m))
		info.append(getattr(ri, m_std))
		
		names.append('mass (ugC)')
		names.append('mass std. (ugC)')

	d13C = 'd13C_' + flag
	d13C_std = 'd13C_' + flag + '_std'

	if hasattr(ri, d13C):
		
		info.append(getattr(ri, d13C))
		info.append(getattr(ri, d13C_std))
		
		names.append('d13C (VPDB)')
		names.append('d13C std. (VPDB)')

	Fm = 'Fm_' + flag
	Fm_std = 'Fm_' + flag + '_std'

	if hasattr(ri, Fm):
		
		info.append(getattr(ri, Fm))
		info.append(getattr(ri, Fm_std))
		
		names.append('Fm')
		names.append('Fm std.')

	info = np.column_stack(info)
	
	# REVISION NOTES:
	#	2023.05.29: CHANGED TO 'display.precison' FOR PANDAS V 2.0
	
	#set pandas display options
	pd.set_option('display.precision', 2)

	#store in dataframe
	frac_info = pd.DataFrame(
		info,
		columns = names,
		index = np.arange(1, ri.nFrac + 1))

	return frac_info


'''
This module contains the Model superclass and all corresponding subclasses.
'''

from __future__ import(
	division,
	print_function,
	)

__docformat__ = 'restructuredtext en'
__all__ = ['Daem']

import matplotlib.pyplot as plt
import numpy as np
import warnings

#import exceptions
from .exceptions import(
	ScalarError,
	)

#import helper functions
from .core_functions import(
	assert_len,
	derivatize,
	)

from .model_helper import(
	_calc_p,
	_bd_calc_A,
	_rpo_calc_A,
	)

class Model(object):
	'''
	Class to store model setup. Intended for subclassing, do not call
	directly.
	'''

	def __init__(self, A, t, T):
		'''
		Initialize the superclass.

		Parameters
		----------
		A : 2d array-like
			Array of the transform matrix to convert from time to rate space.
			Rows are timepoints and columns are k/E values. Shape 
			[`nt` x `nk`].

		t : array-like
			Array of time, in seconds. Length `nt`.

		T : scalar or array-like
			Scalar or array of temperature, in Kelvin. If array, length `nt`.
		'''

		#ensure data is in the right form
		nt = len(t)
		t = assert_len(t, nt)
		T = assert_len(T, nt)
		A = assert_len(A, nt)

		#store attributes
		self.A = A
		self.nt = nt
		self.t = t
		self.T = T

	#define a class method for creating instance directly from timedata
	@classmethod
	def from_timedata(self):
		raise NotImplementedError

	#define a class method for creating instance directly from ratedata
	@classmethod
	def from_ratedata(self):
		raise NotImplementedError

	#define a method for calculating the L curve
	def calc_L_curve(
			self, 
			timedata, 
			ax = None, 
			nLam = 150, 
			lam_max = 1e2, 
			lam_min = 1e-3, 
			plot = False):
		'''
		Function to calculate the L-curve for a given model and timedata
		instance in order to choose the best-fit smoothing parameter, lambda.

		Parameters
		----------
		timedata : rp.TimeData
			``rp.TimeData`` instance containing the time and fraction
			remaining arrays to use in L curve calculation.

		ax : None or matplotlib.axis
			Axis to plot on. If `None` and ``plot = True``, automatically 
			creates a ``matplotlip.axis`` instance to return. Defaults to 
			`None`.

		nLam : int
			Number of lambda values to consider. Defaults to 150.

		lam_max : float or int
			Maximum lambda value to search. Defaults to 1e2.

		lam_min : float or int
			Minimum lambda value to search. Defaults to 1e-3.

		plot : Boolean
			Tells the method to plot the resulting L curve or not. Defaults to
			`False`.

		Returns
		-------
		lam_best : float
			The calculated best-fit lambda value.

		axis : None or matplotlib.axis
			If ``plot = True``, returns an updated axis handle with plot.

		Raises
		------
		ScalarError
			If `lam_max` or `lam_min` are not scalar.

		ScalarError
			If `nLam` is not int.

		See Also
		--------
		calc_L_curve
			Package-level method for ``calc_L_curve``.

		References
		----------
		[1] D.C. Forney and D.H. Rothman (2012) Inverse method for calculating
			respiration rates from decay time series. *Biogeosciences*, **9**,
			3601-3612.

		[2] P.C. Hansen (1994) Regularization tools: A Matlab package for 
			analysis and solution of discrete ill-posed problems. *Numerical*
			*Algorithms*, **6**, 1-35.
		'''

		#check that nLam, lam_max, and lam_min are in the right form
		if not isinstance(lam_max, (int, float)):
			raise ScalarError(
				'lam_max must be float or int')

		elif not isinstance(lam_min, (int, float)):
			raise ScalarError(
				'lam_min must be float or int')

		elif not isinstance(nLam, int):
			raise ScalarError(
				'nLam must be int')

		#define arrays
		log_lam_vec = np.linspace(np.log10(lam_min), np.log10(lam_max), nLam)
		lam_vec = 10**log_lam_vec

		res_vec = np.zeros(nLam)
		rgh_vec = np.zeros(nLam)

		#for each lambda value in the vector, calculate the errors
		for i, w in enumerate(lam_vec):
			_, res, rgh = _calc_p(self, timedata, w)
			res_vec[i] = res
			rgh_vec[i] = rgh

		#convert to log space
		res_vec = np.log10(res_vec)
		rgh_vec = np.log10(rgh_vec)

		#remove noise after 6 sig figs
		res_vec = np.around(res_vec, decimals = 6)
		rgh_vec = np.around(rgh_vec, decimals = 6)

		#calculate derivatives and curvature
		dydx = derivatize(rgh_vec, res_vec)
		dy2d2x = derivatize(dydx, res_vec)

		#function for curvature
		k = dy2d2x/(1+dydx**2)**1.5

		#find first occurrance of argmax k, ignoring first and last points
		i = np.argmax(k[1:-1])
		i += 1 #account for the fact that we dropped the first point
		lam_best = lam_vec[i]

		#plot if necessary
		if plot:

			#create axis if necessary
			if ax is None:
				_,ax = plt.subplots(1, 1)

			#plot results
			ax.plot(
				res_vec,
				rgh_vec,
				linewidth=2,
				color='k',
				label='L-curve')

			ax.scatter(
				res_vec[i],
				rgh_vec[i],
				s=50,
				facecolor='w',
				edgecolor='k',
				linewidth=1.5,
				label=r'best-fit $\lambda$')

			#set axis labels and text

			xlab = r'residual error, $\log_{10} \left( \frac{\|\|' \
				r'\mathbf{A}\cdot \mathbf{p} - \mathbf{g} \|\|}{\sqrt{n_{j}}}' \
				r'\right)$'
			ax.set_xlabel(xlab)
			
			ylab = r'roughness, $\log_{10} \left( \frac{\|\| \mathbf{R}' \
				r'\cdot\mathbf{p} \|\|}{\sqrt{n_{l}}} \right)$'

			ax.set_ylabel(ylab)

			label1 = r'best-fit $\lambda$ = %.3f' %(lam_best)
			
			label2 = (
				r'$log_{10}$ (resid. err.) = %.3f' %(res_vec[i]))
			
			label3 = (
				r'$log_{10}$ (roughness)  = %0.3f' %(rgh_vec[i]))

			ax.text(
				0.5,
				0.95,
				label1 + '\n' + label2 + '\n' + label3,
				verticalalignment='top',
				horizontalalignment='left',
				transform=ax.transAxes)

			#make tight layout
			plt.tight_layout()

			return lam_best, ax

		else:
			return lam_best


class Daem(Model):
	__doc__='''
	Class to calculate the `DAEM` model transform. Used for ramped-temperature
	kinetic problems such as Ramped PyrOx, pyGC, TGA, etc.
	
	Parameters
	----------
	E : array-like
		Array of E values, in kJ/mol. Length `nE`.

	log10omega : scalar, array-like, or lambda function
		Arrhenius pre-exponential factor, either a constant value, array-like
		with length `nE`, or a lambda function of E (in kJ). 

	t : array-like
		Array of time, in seconds. Length `nt`.

	T : array-like
		Array of temperature, in Kelvin. Length `nt`.

	Warnings
	--------
	UserWarning
		If attempting to use isothermal data to create a ``Daem`` model 
		instance.

	See Also
	--------
	RpoThermogram
		``rp.TimeData`` subclass for storing and analyzing RPO 
		time/temperature data.

	EnergyComplex
		``rp.RateData`` subclass for storing and analyzing RPO rate data.

	Examples
	--------
	Creating a DAEM using manually-inputted `E`, `omega`, `t`, and `T`::

		#import modules
		import numpy as np
		import rampedpyrox as rp

		#generate arbitrary data
		t = np.arange(1,100) #100 second experiment
		beta = 0.5 #K/second
		T = beta*t + 273.15 #K
		
		E = np.arange(50, 350) #kJ/mol
		log10omega = 10 #s-1

		#create instance
		daem = rp.Daem(E, log10omega, t, T)

	Creating a DAEM from real thermogram data using the ``rp.Daem.from_timedata``
	class method::

		#import modules
		import rampedpyrox as rp

		#create thermogram instance
		tg = rp.RpoThermogram.from_csv('some_data_file.csv')

		#create Daem instance
		daem = rp.Daem.from_timedata(
			tg, 
			E_max = 350, 
			E_min = 50, 
			nE = 250, 
			log10omega = 10)

	Creating a DAEM from an energy complex using the
	``rp.Daem.from_ratedata`` class method::

		#import modules
		import rampedpyrox as rp

		#create energycomplex instance
		ec = rp.EnergyComplex(E, p0E)

		#create Daem instance
		daem = rp.Daem.from_ratedata(
			ec, 
			beta = 0.08, 
			log10omega = 10, 
			nt = 250, 
			t0 = 0, 
			T0 = 373, 
			tf = 1e4)

	Plotting the L-curve of a Daem to find the best-fit lambda value::

		#import modules
		import matplotlib.pyplot as plt

		#create figure
		fig, ax = plt.subplots(1,1)

		#plot L curve
		lam_best, ax = daem.calc_L_curve(
			tg,
			ax = None, 
			plot = True,
			lam_min = 1e-3,
			lam_max = 1e2,
			nLam = 150)

	**Attributes**

	A : np.ndarray

	E : np.ndarray
		Array of E values, in kJ/mol. Length `nE`.

	nE : int
		Number of activation energy points.

	nt : int
		Number of timepoints.

	t : np.ndarray
		Array of timepoints, in seconds. Length `nt`.

	T : np.ndarray
		Array of temperature, in Kelvin. Length `nt`.

	References
	----------
	[1] R.L Braun and A.K. Burnham (1987) Analysis of chemical reaction 
		kinetics using a distribution of activation energies and simpler 
		models. *Energy & Fuels*, **1**, 153-161.

	[2] B. Cramer (2004) Methane generation from coal during open system
		pyrolysis investigated by isotope specific, Gaussian distributed
		reaction kinetics. *Organic Geochemistry*, **35**, 279-392.

	[3] V. Dieckmann (2005) Modeling petroleum formation from heterogeneous
		source rocks: The influence of frequency factors on activation energy
		distribution and geological prediction. *Marine and Petroleum*
		*Geology*, **22**, 375-390.

	[4] D.C. Forney and D.H. Rothman (2012) Common structure in the
		heterogeneity of plant-matter decay. *Journal of the Royal Society*
		*Interface*, rsif.2012.0122.

	[5] D.C. Forney and D.H. Rothman (2012) Inverse method for calculating
		respiration rates from decay time series. *Biogeosciences*, **9**,
		3601-3612.

	[6] P.C. Hansen (1994) Regularization tools: A Matlab package for analysis
		and solution of discrete ill-posed problems. *Numerical Algorithms*, 
		**6**, 1-35.

	[7] C.C. Lakshmananan et al. (1991) Implications of multiplicity in
		kinetic parameters to petroleum exploration: Distributed activation
		energy models. *Energy & Fuels*, **5**, 110-117.

	[8] J.E. White et al. (2011) Biomass pyrolysis kinetics: A comparative
		critical review with relevant agricultural residue case studies.
		*Journal of Analytical and Applied Pyrolysis*, **91**, 1-33.
	'''

	def __init__(self, E, log10omega, t, T):

		#warn if T is scalar
		if isinstance(T, (int, float)):
			warnings.warn(
				'Attempting to use isothermal data for RPO run! T is a scalar'
				' value of: %r. Consider using an isothermal model type'
				' instead.' % T, UserWarning)

		elif len(set(T)) == 1:
			warnings.warn(
				'Attempting to use isothermal data for RPO run! T is a scalar'
				' value of: %r. Consider using an isothermal model type'
				' instead.' % T[0], UserWarning)

		#get log10omega into the right format
		if hasattr(log10omega,'__call__'):
			log10omega = log10omega(E)

		#calculate A matrix
		A = _rpo_calc_A(E, log10omega, t, T)

		super(Daem, self).__init__(A, t, T)

		#store Daem-specific attributes
		nE = len(E)
		self.log10omega = assert_len(log10omega, nE)
		self.E = assert_len(E, nE)
		self.nE = nE

	@classmethod
	def from_timedata(
			cls, 
			timedata, 
			E_max = 350, 
			E_min = 50, 
			log10omega = 10, 
			nE = 250):
		'''
		Class method to directly generate an ``rp.Daem`` instance using data
		stored in an ``rp.TimeData`` instance.

		Parameters
		----------
		timedata : rp.TimeData
			``rp.TimeData`` instance containing the time array to use
			for creating the DAEM.

		E_max : int
			The maximum activation energy value to consider, in kJ/mol.
			Defaults to 350.

		E_min : int
			The minimum activation energy value to consider, in kJ/mol.
			Defaults to 50.

		log10omega : scalar, array-like, or lambda function
			Arrhenius pre-exponential factor, either a constant value, array-
			likewith length `nE`, or a lambda function of E. Defaults to 10.
		
		nE : int
			The number of activation energy points. Defaults to 250.

		Warnings
		--------
		UserWarning
			If attempting to create a DAEM with an isothermal timedata 
			instance.

		See Also
		--------
		from_ratedata
			Class method to directly generate an ``rp.Daem`` instance using 
			data stored in an ``rp.RateData`` instance.
		'''

		#warn if timedata is not RpoThermogram
		td_type = type(timedata).__name__

		if td_type not in ['RpoThermogram']:
			warnings.warn(
				'Attempting to calculate p distribution using an isothermal'
				' timedata instance of type %r. Consider using rp.RpoThermogram' 
				' instance instead' % td_type, UserWarning)

		#generate E, t, and T array
		E = np.linspace(E_min, E_max, nE)
		t = timedata.t
		T = timedata.T

		return cls(E, log10omega, t, T)

	@classmethod
	def from_ratedata(
			cls, 
			ratedata, 
			beta = 0.08, 
			log10omega = 10, 
			nt = 250,
			t0 = 0, 
			T0 = 373, 
			tf = 1e4):
		'''
		Class method to directly generate an ``rp.Daem`` instance using data
		stored in an ``rp.RateData`` instance.

		Parameters
		----------
		ratedata : rp.RateData
			``rp.RateData`` instance containing the E array to use for
			creating the DAEM. 

		beta : int or float
			Temperature ramp rate to use in model, in Kelvin/second. Defaults
			to 0.08 (*i.e.* 5K/min)

		log10omega : scalar, array-like, or lambda function
			Arrhenius pre-exponential factor, either a constant value, array-
			likewith length `nE`, or a lambda function of E. Defaults to 10.

		nt : int
			The number of time points to use. Defaults to 250.

		t0 : int or float
			The initial time to be used in the model, in seconds. Defaults 
			to 0.

		T0 : int or float
			The initial temperature to be used in the model, in Kelvin.
			Defaults to 373.

		tf : int or float
			The final time to be used in the model, in seconds. Defaults to
			10,000.

		Warnings
		--------
		UserWarning
			If attempting to create a DAEM with a non-EnergyComplex ratedata 
			instance.

		See Also
		--------
		from_timedata
			Class method to directly generate an ``rp.Daem`` instance using data
			stored in an ``rp.TimeData`` instance.

		'''

		#warn if ratedata is not EnergyComplex
		rd_type = type(ratedata).__name__

		if rd_type not in ['EnergyComplex']:
			warnings.warn(
				'Attempting to calculate isotopes using a ratedata instance of'
				' type %r. Consider using rp.EnergyComplex instance instead'
				% rd_type, UserWarning)

		#generate E, t, and T array
		E = ratedata.E
		t = np.linspace(t0, tf, nt)
		T = T0 + beta*t

		return cls(E, log10omega, t, T)


class LaplaceTransform(Model):
	__doc__='''
	
	ADD DOCSTRING!
	
	'''

	def __init__(self, k, t, T, logged = False):

		#calculate A matrix
		A = _bd_calc_A(k, t, logged = logged)

		super(LaplaceTransform, self).__init__(A, t, T)

		#store LaplaceTransform-specific attributes
		nk = len(k)
		self.k = assert_len(k, nk)
		self.nk = nk
		self.logged = logged

	@classmethod
	def from_timedata(
			cls, 
			timedata, 
			k_max = 1, 
			k_min = 1e-6, 
			nk = 250,
			logged = False):
		'''
		Class method to directly generate an ``rp.LaplaceTransform`` instance 
		using data stored in an ``rp.TimeData`` instance.

		Parameters
		----------
		timedata : rp.TimeData
			``rp.TimeData`` instance containing the time array to use
			for creating the LaplaceTransform model.

		k_max : int
			The maximum first-order rate constant value to consider, in s-1 if
			not `logged` is `False` or in ln(s-1) if `logged` is `True`.
			Defaults to 1.

		k_min : int
			The minimum first-order rate constant value to consider, in s-1 if
			not `logged` is `False` or in ln(s-1) if `logged` is `True`.
			Defaults to 1e-6.
		
		nk : int
			The number of rate constant points. Defaults to 250.

		logged : boolean
			If `True`, treats inputted `k` array as the natural log of k (i.e. 
			lambda in Forney and Rothman, 2012 notation). If `False`, treats
			`k` as a linear array. Defaults to `False`.

		Warnings
		--------
		UserWarning
			If attempting to create a LaplaceTransform with a non-isothermal 
			timedata instance.

		See Also
		--------
		from_ratedata
			Class method to directly generate an ``rp.LaplaceTransform`` 
			instance using data stored in an ``rp.RateData`` instance.
		'''

		#warn if timedata is not RpoThermogram
		td_type = type(timedata).__name__

		if td_type not in ['BioDecay']:
			warnings.warn(
				'Attempting to calculate p distribution using a non-isothermal'
				' timedata instance of type %r. Consider using rp.BioDecay' 
				' instance instead' % td_type, UserWarning)

		#generate k, t, and T array
		k = np.linspace(k_min, k_max, nk)
		t = timedata.t
		T = timedata.T

		return cls(k, t, T, logged = logged)

	@classmethod
	def from_ratedata(
			cls, 
			ratedata,
			nt = 250,
			t0 = 0,
			tf = 1e5,
			T = 298):
		'''
		Class method to directly generate an ``rp.LaplaceTransform`` instance 
		using data stored in an ``rp.RateData`` instance.

		Parameters
		----------
		ratedata : rp.RateData
			``rp.RateData`` instance containing the k array to use for
			creating the LaplaceTransform. 

		nt : int
			The number of time points to use. Defaults to 250.

		t0 : int or float
			The initial time to be used in the model, in seconds. Defaults 
			to 0.

		tf : int or float
			The final time to be used in the model, in seconds. Defaults to
			100,000.

		T : int or float
			The temperature of the experiment, in Kelvin. Defaults to 298.

		Warnings
		--------
		UserWarning
			If attempting to create a LaplaceTransform with a non-kDistribution 
			ratedata instance.

		See Also
		--------
		from_timedata
			Class method to directly generate an ``rp.LaplaceTransform`` 
			instance using data stored in an ``rp.TimeData`` instance.
		'''

		#warn if ratedata is not EnergyComplex
		rd_type = type(ratedata).__name__

		if rd_type not in ['kDistribution']:
			warnings.warn(
				'Attempting to calculate isotopes using a ratedata instance of'
				' type %r. Consider using rp.kDistribution instance instead'
				% rd_type, UserWarning)

		#generate k, t, and T array
		k = ratedata.k
		t = np.linspace(t0, tf, nt)
		l = ratedata.logged #make sure this gets stored!

		return cls(k, t, T, logged = l)

if __name__ == '__main__':

	import rampedpyrox as rp

